from typing import Dict, List, Union, Tuple

import numpy as np

from reki.format.grib.common import convert_parameter


def _fill_parameter(
        parameter: Union[str, Dict],
        filter_by_keys: Dict,
        read_keys: List
) -> Tuple[Dict, List]:
    parameter = convert_parameter(parameter)

    if isinstance(parameter, str):
        filter_by_keys["shortName"] = parameter
    elif isinstance(parameter, Dict):
        # TODO: convert int64 to int, move to another place.
        for key, value in parameter.items():
            if isinstance(value, np.int64):
                parameter[key] = int(value)

        filter_by_keys.update(parameter)
        read_keys.extend(parameter.keys())
    else:
        raise ValueError(f"parameter is not supported: {parameter}")
    return filter_by_keys, read_keys


def _fill_level(
        level_type: str,
        level: int,
        filter_by_keys: Dict,
        read_keys: List
) -> Tuple[Dict, List]:
    filter_by_keys.update({
        "typeOfLevel": level_type,
        "level": level,
    })
    return filter_by_keys, read_keys


def _fill_level_type(
        level_type: Union[str, dict],
        filter_by_keys: Dict,
        read_keys: List
) -> Tuple[Dict, List]:
    if isinstance(level_type, dict):
        filter_by_keys.update(level_type)
        read_keys.extend([key for key in level_type if key not in read_keys])
    else:
        filter_by_keys.update({
            "typeOfLevel": level_type,
        })
    return filter_by_keys, read_keys


def _fill_level_value(
        level: Union[int, float],
        filter_by_keys: Dict,
        read_keys: List
) -> Tuple[Dict, List]:
    filter_by_keys.update({
        "level": level,
    })
    return filter_by_keys, read_keys


def _fill_index_path(
        with_index: Union[str, bool],
        backend_kwargs: Dict,
) -> Dict:
    if isinstance(with_index, str):
        backend_kwargs["indexpath"] = with_index
    elif isinstance(with_index, bool):
        if not with_index:
            backend_kwargs["indexpath"] = ""
    return backend_kwargs



from typing import Union, Literal

import xarray as xr

from ._interpolator import _get_interpolator


def interpolate_grid(
        data: xr.DataArray,
        target: xr.DataArray,
        scheme: str = "linear",
        engine: Literal["scipy", "xarray"] = "xarray",
        **kwargs
) -> xr.DataArray:
    """
    Interpolate grid data into a target grid.

    Parameters
    ----------
    data : xr.DataArray
        intput data
    target : xr.DataArray
        target grid
    scheme : str
        interpolate method.

        * if ``engine="xarray"``, `linear` or `nearest`
        * if ``engine="scipy"``, `linear`, `nearest`, `splinef2d` or `rect_bivariate_spline`
    engine : str
        interpolate engine, `xarray` or `scipy`
    kwargs
        key-value parameters to be passed to interpolator with _get_interpolator function.

    Returns
    -------
    xr.DataArray
    """
    interpolator = _get_interpolator(scheme, engine, **kwargs)

    target_field = interpolator.interpolate_grid(
        data=data,
        target=target,
    )

    return target_field


def extract_point(
        data: xr.DataArray,
        latitude: Union[float, int, list[Union[float, int]]],
        longitude: Union[float, int, list[Union[float, int]]],
        scheme: str = "linear",
        engine: Literal["scipy", "xarray"] = "xarray",
        **kwargs
) -> xr.DataArray:
    """
    Extract a point from 2D field with interpolation.

    Parameters
    ----------
    data
    latitude
    longitude
    scheme
        interpolate method.

        * if ``engine="xarray"``, `linear` or `nearest`
        * if ``engine="scipy"``, `linear`, `nearest`, `splinef2d` or `rect_bivariate_spline`
    engine
        interpolate engine, `xarray` or `scipy`.
    kwargs

    Returns
    -------
    xr.DataArray
    """
    interpolator = _get_interpolator(scheme, engine, **kwargs)
    value = interpolator.extract_point(data, latitude=latitude, longitude=longitude)
    return value



from typing import Union, Optional, Literal
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr


from .grads_ctl import GradsCtlParser
from .grads_data_handler import GradsDataHandler, GradsRecordHandler


def load_field_from_file(
        file_path: Union[str, Path],
        parameter: str,
        level_type: Optional[str] = None,
        level: Union[int, float, list] = None,
        level_dim: Optional[str] = None,
        latitude_direction: Literal["degree_north", "degree_south"] = "degree_north",
        valid_time: Union[str, pd.Timestamp] = None,
        forecast_time: Union[str, pd.Timedelta] = None,
        **kwargs
) -> Optional[xr.DataArray]:
    """
    Load one field or fields of one parameter from GrADS binary file.

    Parameters
    ----------
    file_path
        CTL description file path.
    parameter : str
        parameter name
    level_type : Optional[str]
        type of level

        * pl / ml
        * index
        * single
        * None
    level : Union[int, float, list]
        level value
    level_dim
    latitude_direction
        * degree_north
        * degree_south
    valid_time
    forecast_time
    kwargs

    Returns
    -------
    xr.DataArray
        Xarray DataArray if found, or None if not.

    Examples
    --------
    Get CMA-GFS GMF modelvar data file path.

    >>> from reki.data_finder import find_local_file
    >>> from reki.format.grads import load_field_from_file
    >>> modelvar_file_path = find_local_file(
    ...     "cma_gfs_gmf/bin/modelvar_ctl",
    ...     start_time="2024102600",
    ...     forecast_time="36h"
    ... )
    >>> modelvar_file_path
    PosixPath('/g3/COMMONDATA/OPER/CEMC/GFS_GMF/Fcst-long/2024102600/model.ctl_2024102600_036')

    Load a single level field:

    >>> field = load_field_from_file(
    ...     modelvar_file_path,
    ...     parameter="ts",
    ...     level_type="single",
    ... )
    >>> field
    <xarray.DataArray 'ts' (latitude: 1671, longitude: 2501)> Size: 17MB
    array([[276.65912, 276.66272, 276.66632, ..., 265.26233, 265.36075,
            265.45914],
           [271.72498, 271.74908, 272.00296, ..., 258.56754, 258.8194 ,
            265.57312],
           [271.73434, 271.7566 , 272.02072, ..., 259.0332 , 258.90613,
            265.68716],
           ...,
           [302.55112, 302.55115, 302.55112, ..., 302.75116, 302.75116,
            302.75116],
           [302.5451 , 302.5451 , 302.5451 , ..., 302.74515, 302.74515,
            302.74515],
           [302.53912, 302.53912, 302.53912, ..., 302.73914, 302.73914,
            302.73914]], dtype=float32)
    Coordinates:
      * latitude    (latitude) float64 13kB 60.1 60.07 60.04 ... 10.06 10.03 10.0
      * longitude   (longitude) float64 20kB 70.0 70.03 70.06 ... 144.9 145.0 145.0
        level       int64 8B 0
        valid_time  datetime64[ns] 8B ...
    Attributes:
        description:  surface temperature

    Load a multi level field:

    >>> field = load_field_from_file(
    ...     modelvar_file_path,
    ...     parameter="pi",
    ...     level_type="ml",
    ...     level=[1,2,3,4]
    ... )
    >>> field
    <xarray.DataArray 'pi' (ml: 4, latitude: 1441, longitude: 2880)> Size: 66MB
    array([[[0.9994781 , 0.99947816, 0.9994783 , ..., 0.99947804,
             0.9994781 , 0.99947816],
            [0.9994006 , 0.99940056, 0.9994005 , ..., 0.99940073,
             0.9994006 , 0.9994006 ],
            [0.99941564, 0.9994156 , 0.9994156 , ..., 0.99941593,
             0.9994159 , 0.99941576],
            ...,
            [0.9025082 , 0.90250385, 0.90249956, ..., 0.902521  ,
             0.9025168 , 0.9025125 ],
            [0.9023999 , 0.90239733, 0.9023947 , ..., 0.90240735,
             0.9024049 , 0.90240246],
            [0.90238315, 0.90238225, 0.90238154, ..., 0.9023854 ,
             0.9023846 , 0.9023839 ]],
           [[0.9987454 , 0.99874544, 0.99874544, ..., 0.9987453 ,
             0.9987453 , 0.9987454 ],
            [0.9986675 , 0.9986674 , 0.99866736, ..., 0.9986676 ,
             0.99866754, 0.9986675 ],
            [0.99867845, 0.9986784 , 0.9986783 , ..., 0.99867874,
             0.9986786 , 0.99867857],
    ...
            [0.90097326, 0.90096897, 0.9009646 , ..., 0.90098614,
             0.90098184, 0.90097755],
            [0.9008647 , 0.9008622 , 0.90085965, ..., 0.9008723 ,
             0.9008698 , 0.9008672 ],
            [0.9008468 , 0.900846  , 0.9008453 , ..., 0.90084904,
             0.90084827, 0.90084755]],
           [[0.99662673, 0.9966268 , 0.99662685, ..., 0.9966266 ,
             0.9966266 , 0.9966267 ],
            [0.996547  , 0.9965469 , 0.9965469 , ..., 0.9965471 ,
             0.99654704, 0.99654704],
            [0.9965461 , 0.99654603, 0.99654603, ..., 0.99654627,
             0.9965462 , 0.99654615],
            ...,
            [0.8998071 , 0.89980286, 0.8997985 , ..., 0.8998199 ,
             0.8998157 , 0.89981145],
            [0.89969814, 0.8996955 , 0.899693  , ..., 0.89970565,
             0.89970315, 0.89970064],
            [0.8996787 , 0.89967805, 0.8996773 , ..., 0.89968103,
             0.89968026, 0.8996795 ]]], dtype=float32)
    Coordinates:
      * latitude       (latitude) float64 12kB 90.0 89.88 89.75 ... -89.88 -90.0
      * longitude      (longitude) float64 23kB 0.0 0.125 0.25 ... 359.6 359.8 359.9
      * ml             (ml) float64 32B 1.0 2.0 3.0 4.0
        valid_time     datetime64[ns] 8B ...
        start_time     datetime64[ns] 8B ...
        forecast_time  timedelta64[ns] 8B 1 days 12:00:00
    Attributes:
        description:  PI

    Get CMA-MESO postvar data file path.

    >>> postvar_file_path = find_local_file(
    ...     "cma_meso_3km/bin/postvar_ctl",
    ...     start_time="2024102600",
    ... )
    >>> postvar_file_path
    PosixPath('/g3/COMMONDATA/OPER/CEMC/MESO_3KM/Fcst-main/2024102600/postvar.ctl_202410260000000')

    Load a field. Set ``forecast_time`` option because all data files of one cycle use one CTL file.

    >>> field = load_field_from_file(
    ...     postvar_file_path,
    ...     parameter="t",
    ...     level_type="pl",
    ...     forecast_time="24h",
    ... )
    >>> field
    <xarray.DataArray 't' (pl: 26, latitude: 1671, longitude: 2501)> Size: 435MB
    array([[[275.32828, 275.32657, 275.3065 , ..., 260.2067 , 260.09564,
             259.9529 ],
            [275.34738, 275.35904, 275.31445, ..., 260.0906 , 259.94995,
             259.89807],
            [275.40707, 275.39713, 275.38556, ..., 260.17816, 259.95865,
             259.75906],
            ...,
            [299.74286, 299.75177, 299.7522 , ..., 300.81787, 300.75705,
             300.7689 ],
            [299.73053, 299.7287 , 299.73077, ..., 300.78677, 300.80746,
             300.80475],
            [299.73132, 299.7241 , 299.72073, ..., 300.7902 , 300.8348 ,
             300.81497]],
           [[274.36758, 274.35873, 274.33942, ..., 258.95673, 258.8462 ,
             258.70416],
            [274.3802 , 274.38016, 274.39774, ..., 258.84058, 258.70056,
             258.64957],
            [274.4292 , 274.4326 , 274.42468, ..., 258.92773, 258.70926,
             258.51123],
    ...
            [221.9119 , 221.9043 , 221.90732, ..., 220.80681, 220.79927,
             220.79141],
            [221.9006 , 221.89711, 221.89825, ..., 220.78723, 220.77835,
             220.76967],
            [221.88934, 221.88995, 221.89085, ..., 220.76501, 220.7564 ,
             220.74776]],
           [[218.22253, 218.22556, 218.22949, ..., 225.40901, 225.4072 ,
             225.40222],
            [218.22551, 218.2267 , 218.22943, ..., 225.07826, 225.0852 ,
             225.40283],
            [218.2302 , 218.23157, 218.23302, ..., 225.08052, 225.0856 ,
             225.39835],
            ...,
            [234.57108, 234.70776, 234.7125 , ..., 232.78271, 232.81221,
             232.84262],
            [234.57117, 234.64745, 234.65125, ..., 232.78842, 232.81712,
             232.84677],
            [234.57133, 234.57503, 234.57837, ..., 232.82465, 232.83774,
             232.85094]]], dtype=float32)
    Coordinates:
      * latitude    (latitude) float64 13kB 60.1 60.07 60.04 ... 10.06 10.03 10.0
      * longitude   (longitude) float64 20kB 70.0 70.03 70.06 ... 144.9 145.0 145.0
      * pl          (pl) float64 208B 1e+03 975.0 950.0 925.0 ... 30.0 20.0 10.0
        valid_time  datetime64[ns] 8B ...
    Attributes:
        description:  temperature
    """
    if isinstance(forecast_time, str):
        forecast_time = pd.to_timedelta(forecast_time)
    if isinstance(file_path, str):
        file_path = Path(file_path)
    if isinstance(valid_time, str):
        valid_time = pd.to_datetime(valid_time)

    ctl_parser = GradsCtlParser()
    ctl_parser.parse(file_path)
    grads_ctl = ctl_parser.grads_ctl

    # level_type: pl, index, single
    grads_level_type = "multi"
    level_dim_name = "level"

    if not isinstance(level, list) and level is not None:
        level = [level]

    if level_type == "single":
        level = [0]
        grads_level_type = "single"
    elif level_type == "index":
        level = [grads_ctl.zdef["values"][cur_level] for cur_level in level]
    elif level_type in ("pl", "ml"):
        level_dim_name = level_type
        if level is None:
            level = grads_ctl.zdef["values"]
    elif level_type is None:
        grads_level_type = None
        # level = None

    if level_dim is not None:
        level_dim_name = level_dim

    data_handler = GradsDataHandler(grads_ctl)

    xarray_records = []
    for index, record in enumerate(grads_ctl.record):
        if not check_record(
            record,
            parameter=parameter,
            level=level,
            level_type=grads_level_type,
            valid_time=valid_time,
            forecast_time=forecast_time,
        ):
            continue

        offset = data_handler.get_offset_by_record_index(record["record_index"])
        record_handler = GradsRecordHandler(grads_ctl, index, offset)

        xarray_record = create_data_array_from_record(
            record=record_handler,
            parameter=parameter,
            level=record["level"],
            level_dim_name=level_dim_name,
            latitude_direction=latitude_direction,
        )
        xarray_records.append(xarray_record)

    record_count = len(xarray_records)
    if record_count == 0:
        return None
    elif record_count == 1:
        return xarray_records[0]
    else:
        data = xr.concat(xarray_records, level_dim_name)

    return data


def check_record(
        record: dict,
        parameter: str,
        level_type: str = None,
        level: Union[int, float, list] = None,
        valid_time: pd.Timestamp = None,
        forecast_time: pd.Timedelta = None
) -> bool:
    if parameter != record["name"]:
        return False

    if level_type is not None and level_type != record["level_type"]:
        return False

    if level is not None:
        if isinstance(level, list):
            if record["level"] not in level:
                return False
        else:
            if level != record["level"]:
                return False

    if valid_time is not None and valid_time != record["valid_time"]:
        return False

    if forecast_time is not None and forecast_time != record["forecast_time"]:
        return False

    return True


def create_data_array_from_record(
        record: GradsRecordHandler,
        parameter,
        level,
        level_dim_name=None,
        latitude_direction="degree_north",
) -> Optional[xr.DataArray]:
    grads_ctl = record.grads_ctl

    # values
    file_path = grads_ctl.get_data_file_path(record.record_info)
    with open(file_path, "rb") as f:
        values = record.load_data(f)

    # coords
    lons = grads_ctl.xdef["values"]
    lats = grads_ctl.ydef["values"]

    if latitude_direction == "degree_north":
        values = np.flip(values, 0)
        lats = lats[::-1]

    coords = {}
    coords["latitude"] = xr.Variable(
        "latitude",
        lats,
        attrs={
            "units": latitude_direction,
            "standard_name": "latitude",
            "long_name": "latitude"
        },
    )
    coords["longitude"] = xr.Variable(
        "longitude",
        lons,
        attrs={
            "units": "degrees_east",
            "standard_name": "longitude",
            "long_name": "longitude"
        }
    )

    coords[level_dim_name] = level
    coords["valid_time"] = record.record_info["valid_time"]

    if grads_ctl.start_time is not None and grads_ctl.forecast_time is not None:
        coords["start_time"] = grads_ctl.start_time
        coords["forecast_time"] = grads_ctl.forecast_time

    # dims
    dims = ("latitude", "longitude")

    # attrs
    data_attrs = {
        "description": record.record_info["description"]
    }

    data = xr.DataArray(
        values,
        dims=dims,
        coords=coords,
        attrs=data_attrs,
        name=parameter,
    )

    return data



import pytest
from dataclasses import dataclass, asdict

from reki.data_finder import find_local_file
from reki.format.grib.eccodes import load_field_from_file


@dataclass
class QueryOption:
    parameter: str
    level_type: str
    level: float


@dataclass
class TestCase:
    query: QueryOption


@pytest.fixture
def system_name():
    return "cma_meso_3km"


@pytest.fixture
def file_path(system_name, start_time, forecast_time, storage_base):
    f = find_local_file(
        f"{system_name}/grib2/orig",
        start_time=start_time,
        forecast_time=forecast_time,
        storage_base=storage_base
    )
    return f


@pytest.fixture
def modelvar_file_path(system_name, start_time, forecast_time, storage_base):
    f = find_local_file(
        f"{system_name}/grib2/modelvar",
        start_time=start_time,
        forecast_time=forecast_time,
        storage_base=storage_base
    )
    return f


def test_load_with_short_name(file_path):
    """
    Use short name supported by ecCodes
    """
    short_name = "t"
    field = load_field_from_file(
        file_path,
        parameter=short_name,
        level_type="pl",
        level=850
    )
    assert field is not None


def test_load_with_extended_name(file_path):
    """
    Use short name embedded by reki from wgrib2 and CEMC
    """
    parameter = "TCOLW"
    field = load_field_from_file(
        file_path,
        parameter=parameter,
    )
    assert field is not None


def test_load_with_numbers(file_path):
    """
    Use GRIB keys for parameter:

    * discipline
    * parameterCategory
    * parameterNumber
    """
    parameter = {
        "discipline": 0,
        "parameterCategory": 1,
        "parameterNumber": 225
    }
    level_type = "pl"
    level = 850
    field = load_field_from_file(
        file_path,
        parameter=parameter,
        level_type=level_type,
        level=level
    )
    assert field is not None


def test_load_wind_10m(file_path):
    """
    Use ``stepType`` for normal and statistics fields.
    """
    parameter = "10u"

    # 10m U 风
    step_type = "instant"
    field_instant = load_field_from_file(
        file_path,
        parameter=parameter,
        stepType=step_type
    )
    assert field_instant is not None

    # 输出间隔内最大 10m U 风
    step_type = "max"
    field_max = load_field_from_file(
        file_path,
        parameter=parameter,
        stepType=step_type
    )
    assert field_max is not None


def test_load_ri(file_path):
    """
    Use different ``level_type`` for same parameter.

    地表理查森数
    边界层理查森数
    """
    parameter = "RI"
    level_type = "surface"
    field_surface = load_field_from_file(
        file_path,
        parameter=parameter,
        level_type=level_type
    )
    assert field_surface is not None

    level_type = {
        "typeOfFirstFixedSurface": 166
    }
    field_166 = load_field_from_file(
        file_path,
        parameter=parameter,
        level_type=level_type
    )
    assert field_166 is not None


def test_load_with_two_levels(file_path):
    """
    Use ``first_level`` and ``second_level`` in ``level`` option.

    土壤温度

    * 0-10cm below ground
    * 10-30cm below ground
    """
    parameter = "t"
    level_type = "depthBelowLandLayer"
    first_level = 0
    second_level = 0.1

    field_0_10 = load_field_from_file(
        file_path,
        parameter=parameter,
        level_type=level_type,
        level={
            "first_level": first_level,
            "second_level": second_level
        }
    )
    assert field_0_10 is not None

    first_level = 0.1
    second_level = 0.3

    field_10_30 = load_field_from_file(
        file_path,
        parameter=parameter,
        level_type=level_type,
        level={
            "first_level": first_level,
            "second_level": second_level
        }
    )
    assert field_10_30 is not None


def test_load_modelvar_using_level_type(modelvar_file_path):
    parameter = "t"
    level = 10
    level_type = "ml"
    field = load_field_from_file(
        modelvar_file_path,
        parameter=parameter,
        level_type=level_type,
        level=level
    )
    assert field is not None


def test_load_modelvar_using_cemc_param_table(modelvar_file_path):
    test_cases = [
        TestCase(QueryOption(parameter="pip", level_type="ml", level=30)),
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            modelvar_file_path,
            **asdict(test_case.query)
        )
        assert field is not None



import pytest
import pandas as pd

from reki.data_finder import find_local_file


@pytest.fixture
def storage_base():
    return "/mnt/m/archive"


@pytest.fixture
def start_time():
    return pd.Timestamp.now().ceil(freq="D") - pd.Timedelta(days=2)


@pytest.fixture
def forecast_time():
    return pd.to_timedelta("24h")


@pytest.fixture
def system_name():
    return "cma_meso_3km"


@pytest.fixture
def file_path(system_name, start_time, storage_base):
    f = find_local_file(
        f"{system_name}/bin/postvar_ctl",
        start_time=start_time,
        storage_base=storage_base
    )
    return f


@pytest.fixture
def modelvar_file_path(system_name, start_time, storage_base):
    f = find_local_file(
        f"{system_name}/bin/modelvar_ctl",
        start_time=start_time,
        storage_base=storage_base
    )
    return f



from typing import Union, Dict

from reki.format.grib.config import WGRIB2_SHORT_NAME_TABLE, CEMC_PARAM_TABLE


def convert_parameter(parameter: Union[str, Dict]) -> Union[str, Dict]:
    """
    Convert string parameter into GRIB keys according to short name tables.
    If parmeter is found in tables, it will be replaced by GRIB key dict.
    Or if parameter is not found, return the string.

    Currentlly reki has two short name tables (and search by order):

    * WGRIB short name table
    * CEMC param table

    Parameters
    ----------
    parameter

    Returns
    -------
    Union[str, Dict]

    Examples
    --------
    >>> from reki.format.grib.common._parameter import convert_parameter

    Convert wgrib2 short names:

    >>> convert_parameter("TMP")
    {'discipline': 0, 'parameterCategory': 0, 'parameterNumber': 0}
    >>> convert_parameter("VIS")
    {'discipline': 0, 'parameterCategory': 19, 'parameterNumber': 0}

    Convert CEMC params:

    >>> convert_parameter("bli")
    {'discipline': 0.0, 'parameterCategory': 7.0, 'parameterNumber': 1.0}
    >>> convert_parameter("t2m")
    {'discipline': 0.0, 'parameterCategory': 0.0, 'parameterNumber': 0.0}

    Unknown parameter:

    >>> convert_parameter("unknown")
    'unknown'

    dict parameter:

    >>> convert_parameter({"parameterCategory": 0, "parameterNumber": 0})
    {'parameterCategory': 0, 'parameterNumber': 0}

    """
    if isinstance(parameter, str):
        df = WGRIB2_SHORT_NAME_TABLE[WGRIB2_SHORT_NAME_TABLE["short_name"] == parameter]
        if not df.empty:
            return df.iloc[0].drop("short_name").to_dict()

        param_df = CEMC_PARAM_TABLE[CEMC_PARAM_TABLE["name"] == parameter]
        if not param_df.empty:
            row = param_df.iloc[0]
            return dict(
                discipline=float(row["discipline"]),
                parameterCategory=float(row["category"]),
                parameterNumber=float(row["number"]),
            )

    return parameter



from dataclasses import dataclass, asdict
from typing import List, Union, Optional

from reki.format.grads import load_field_from_file


@dataclass
class QueryOption:
    parameter: str
    level_type: Optional[str]
    level: Optional[Union[int, float, List]]


@dataclass
class TestCase:
    query: QueryOption


def test_string(file_path, forecast_time):
    test_cases = [
        TestCase(query=QueryOption("t", "pl", 850)),
        TestCase(query=QueryOption("u", "pl", 10)),
        TestCase(query=QueryOption("cr", None, None)),
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            forecast_time=forecast_time,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.name == test_case.query.parameter



from typing import Union, Dict, List, Optional
import math

import eccodes


def _check_message(
        message_id,
        parameter: Optional[Union[str, Dict]],
        level_type: Optional[Union[str, List[str]]],
        level: Optional[Union[int, List[int], Dict]],
        **kwargs,
) -> bool:
    """
    Check whether GRIB message (`message_id`) fits conditions.

    Currently, conditions are:
        - parameter
        - level_Type
        - level
        - other GRIB keys

    Parameters
    ----------
    message_id
    parameter
    level_type
    level
    kwargs

    Returns
    -------

    """
    if not _check_parameter(message_id, parameter):
        return False
    if not _check_level_type(message_id, level_type):
        return False
    if not _check_level_value(message_id, level, level_type):
        return False
    if not _check_keys(message_id, **kwargs):
        return False
    return True


def _check_parameter(message_id, parameter: Optional[Union[str, Dict]]) -> bool:
    if parameter is None:
        return True
    # parameter = _convert_parameter(parameter)
    if isinstance(parameter, str):
        short_name = eccodes.codes_get(message_id, "shortName")
        return short_name == parameter
    elif isinstance(parameter, Dict):
        for key, value in parameter.items():
            if eccodes.codes_get(message_id, key) != value:
                return False
        return True
    else:
        raise ValueError(f"parameter is not supported: {parameter}")


def _check_level_type(
        message_id,
        level_type: Optional[Union[str, List, Dict]],
) -> bool:
    if level_type is None:
        return True
    message_type = eccodes.codes_get(message_id, "typeOfLevel", ktype=str)
    if isinstance(level_type, str):
        return message_type == level_type
    elif isinstance(level_type, List):
        for cur_level_type in level_type:
            if _check_level_type(message_id, cur_level_type):
                return True
        return False
    elif isinstance(level_type, Dict):
        for key in level_type:
            requested_value = level_type[key]
            value = eccodes.codes_get(message_id, key, ktype=type(requested_value))
            if requested_value != value:
                return False
        return True
    else:
        raise ValueError(f"level_type is not supported: {level_type}")


def _check_level_value(
        message_id,
        level: Optional[Union[int, float, List[int], Dict]],
        level_type: Union[str, Dict] = None
) -> bool:
    if level is None:
        return True

    message_level = eccodes.codes_get(message_id, "level", ktype=float)

    # check for `pl` using unit hPa.
    # WARNING: This may be changed.
    if isinstance(level_type, Dict):
        if "typeOfFirstFixedSurface" in level_type and level_type["typeOfFirstFixedSurface"] == 100:
            level_in_pa = _get_level_value(message_id, "First")
            message_level = level_in_pa / 100.0

    if isinstance(level, int) or isinstance(level, float):
        return message_level == level
    elif isinstance(level, List):
        return message_level in level
    elif isinstance(level, Dict):
        current_level_dict = level.copy()
        if "first_level" in current_level_dict:
            required_first_level = current_level_dict.pop("first_level")
            first_level = _get_level_value(message_id, "First")
            if first_level != float(required_first_level):
                return False
        if "second_level" in current_level_dict:
            required_second_level = current_level_dict.pop("second_level")
            second_level = _get_level_value(message_id, "Second")
            if second_level != float(required_second_level):
                return False
        return _check_keys(message_id, **current_level_dict)
    elif level == "all":
        return True
    else:
        raise ValueError(f"level is not supported: {level}")


def _check_keys(message_id, **kwargs):
    for key, value in kwargs.items():
        v = eccodes.codes_get(message_id, key)
        if value != v:
            return False
    return True


def _get_level_value(message_id, name: str = "First"):
    f = eccodes.codes_get(message_id, f"scaleFactorOf{name}FixedSurface")
    v = eccodes.codes_get(message_id, f"scaledValueOf{name}FixedSurface")
    level = math.pow(10, -1 * f) * v
    return level



import datetime
from typing import Dict, List, Union, Optional, Iterable, Callable
from pathlib import Path

import pandas as pd
from jinja2 import Template


def find_file(
        config: Dict,
        data_level: Union[str, List],
        start_time: Union[datetime.datetime, pd.Timestamp],
        forecast_time: pd.Timedelta,
        obs_time: Optional[pd.Timedelta] = None,
        debug: bool = False,
        **kwargs
) -> Optional[Path]:
    """
    Find a file according to config.

    Parameters
    ----------
    config
        data source config dictionary. An example of config file:

        .. code-block:: yaml

            query:
              system: cma_gfs_gmf
              stream: oper
              type: grib2
              name: modelvar

            file_name: 'modelvar{{ time_vars.Year }}{{ time_vars.Month }}{{ time_vars.Day }}{{ time_vars.Hour }}{{ time_vars.Forecast }}.grb2'

            paths:
              - type: local
                level: runtime
                path: "/some/path/to/{{ time_vars.Year }}{{ time_vars.Month }}{{ time_vars.Day }}{{ time_vars.Hour }}/data/output/grib2_orig/"

              - type: local
                level: storage
                path: '{{ query_vars.storage_base }}/GRAPES_GFS_GMF/Prod-grib/{{ time_vars.Year }}{{ time_vars.Month }}{{ time_vars.Day }}{{ time_vars.Hour }}/MODELVAR'

    data_level
    start_time
    forecast_time
    obs_time
    debug
        show debug info
    kwargs

    Returns
    -------
    Path if found, None otherwise
    """
    query_vars = QueryVars()

    for key in config["query"]:
        setattr(query_vars, key, config["query"][key])
    for key in kwargs:
        setattr(query_vars, key, kwargs[key])

    time_vars = TimeVars(start_time=start_time, forecast_time=forecast_time)
    if obs_time is not None:
        obs_time_vars = TimeVars(start_time=obs_time)
        setattr(query_vars, "obs_time", obs_time_vars)

    parse_template = generate_template_parser(time_vars, query_vars)
    file_name = parse_template(config["file_name"])
    if debug:
        print("file name:", file_name)
    file_path = None
    paths = config["paths"]
    for a_path_object in paths:
        current_data_level = a_path_object["level"]
        if not check_data_level(current_data_level, data_level):
            continue

        path_template = a_path_object["path"]
        current_dir_path = parse_template(path_template)
        current_file_path = Path(current_dir_path, file_name)
        if debug:
            print("searching file path:", current_file_path)
        if current_file_path.is_file():
            file_path = current_file_path
            break

    return file_path


def render_file_name(
        config: dict,
        start_time: Union[datetime.datetime, pd.Timestamp],
        forecast_time: Union[pd.Timedelta, str],
        obs_time: Optional[pd.Timedelta] = None,
        **kwargs
):
    query_vars = QueryVars()

    for key in config["query"]:
        setattr(query_vars, key, config["query"][key])
    for key in kwargs:
        setattr(query_vars, key, kwargs[key])

    time_vars = TimeVars(start_time=start_time, forecast_time=forecast_time)
    if obs_time is not None:
        obs_time_vars = TimeVars(start_time=obs_time)
        setattr(query_vars, "obs_time", obs_time_vars)

    parse_template = generate_template_parser(time_vars, query_vars)
    if "file_name" in config:
        file_name = parse_template(config["file_name"])
    elif "file_names" in config:
        file_name = parse_template(config["file_names"][0])
    return file_name


def find_files(
        config: dict,
        data_level: Union[str, List],
        start_time: Union[datetime.datetime, pd.Timestamp],
        forecast_time: pd.Timedelta,
        glob: bool = True,
        **kwargs
) -> Optional[List[Path]]:
    query_vars = QueryVars()

    for key in config["query"]:
        setattr(query_vars, key, config["query"][key])
    for key in kwargs:
        setattr(query_vars, key, kwargs[key])

    time_vars = TimeVars(start_time=start_time, forecast_time=forecast_time)

    parse_template = generate_template_parser(time_vars, query_vars)
    file_name = parse_template(config["file_name"])
    file_paths = []
    paths = config["paths"]
    for a_path_object in paths:
        current_data_level = a_path_object["level"]
        if not check_data_level(current_data_level, data_level):
            continue

        path_template = a_path_object["path"]
        current_dir_path = Path(parse_template(path_template))
        current_files = current_dir_path.glob(file_name)
        file_paths.extend([f for f in current_files if f.is_file()])

    if len(file_paths) == 0:
        return None
    return file_paths


def check_data_level(data_level, required_level: Optional[Union[str, Iterable]]) -> bool:
    """
    check whether data level is in required level(s).

    Parameters
    ----------
    data_level
    required_level

    Returns
    -------
    bool
    """
    if required_level is None:
        return True
    elif isinstance(required_level, str):
        return data_level == required_level
    elif isinstance(required_level, Iterable):
        return data_level in required_level
    else:
        raise ValueError(f"level is not supported {required_level}")


def get_hour(forecast_time: pd.Timedelta) -> int:
    return int(forecast_time.seconds/3600) + forecast_time.days * 24


class QueryVars:
    def __init__(self):
        self.storage_base = None


class TimeVars:
    def __init__(
            self,
            start_time: Union[datetime.datetime, pd.Timestamp],
            forecast_time: Union[pd.Timedelta, str] = pd.Timedelta(hours=0)
    ):
        self.Year = start_time.strftime("%Y")
        self.Month = start_time.strftime("%m")
        self.Day = start_time.strftime("%d")
        self.Hour = start_time.strftime("%H")
        self.Minute = start_time.strftime("%M")

        if isinstance(forecast_time, pd.Timedelta):
            self.Forecast = f"{get_hour(forecast_time):03}"
        else:
            self.Forecast = forecast_time

        start_date_time_4dvar = start_time - datetime.timedelta(hours=3)
        self.Year4DV = start_date_time_4dvar.strftime("%Y")
        self.Month4DV = start_date_time_4dvar.strftime("%m")
        self.Day4DV = start_date_time_4dvar.strftime("%d")
        self.Minute4DV = start_time.strftime("%M")
        self.Hour4DV = start_date_time_4dvar.strftime("%H")


def generate_template_parser(time_vars: Dict, query_vars: Dict) -> Callable:

    def parse_template(template_content):
        template = Template(template_content)
        return template.render(time_vars=time_vars, query_vars=query_vars)

    return parse_template



import pytest
import pandas as pd


@pytest.fixture
def cma_gfs_system_name() -> str:
    return "cma_gfs_gmf"


@pytest.fixture
def last_two_day() -> pd.Timestamp:
    current = pd.Timestamp.now().floor(freq="D")
    last_two_day = current - pd.Timedelta(days=2)
    return last_two_day


@pytest.fixture
def forecast_time_24h() -> pd.Timedelta:
    return pd.to_timedelta("24h")


@pytest.fixture
def cma_gfs_data_dir():
    return "/g3/COMMONDATA/OPER/CEMC/GFS_GMF/Prod-grib/{start_time_label}/ORIG"


from dataclasses import dataclass, asdict
from typing import Dict, Union, List, Optional

import pytest
import numpy as np

from reki.format.grib.eccodes import load_field_from_file


@dataclass
class QueryOption:
    parameter: Union[str, Dict] = None
    level_type: Union[str, Dict] = None
    level: Optional[Union[float, str, Dict, List[float]]] = None


@dataclass
class TestCase:
    query: QueryOption
    expected_level_name: str = None
    expected_level: float = None


def test_scalar(file_path, modelvar_file_path):
    test_cases = [
        TestCase(
            query=QueryOption(parameter="t", level_type="pl", level=1.5),
            expected_level_name="pl",
            expected_level=1.5
        ),
        TestCase(
            query=QueryOption(parameter="t", level_type="isobaricInhPa", level=850),
            expected_level_name="isobaricInhPa",
            expected_level=850
        ),
        TestCase(
            query=QueryOption(parameter="TMAX", level_type="heightAboveGround", level=2),
            expected_level_name="heightAboveGround",
            expected_level=2
        ),
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.coords[test_case.expected_level_name] == test_case.expected_level

    test_cases = [
        TestCase(
            query=QueryOption(
                parameter="u", level_type={"typeOfFirstFixedSurface": 131}, level=10),
            expected_level_name="level_131",
            expected_level=10
        )
    ]
    for test_case in test_cases:
        field = load_field_from_file(
            modelvar_file_path,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.coords[test_case.expected_level_name] == test_case.expected_level


def test_dict(file_path):
    test_cases = [
        TestCase(
            QueryOption(
                parameter="vwsh",
                level_type="heightAboveGroundLayer",
                level=dict(first_level=1000, second_level=0)
            ),
        ),
        TestCase(
            QueryOption(
                parameter="t",
                level_type="depthBelowLandLayer",
                level=dict(first_level=0.1, second_level=0.4)
            ),
        )
    ]
    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert field is not None


def test_multi_levels(file_path):
    test_cases = [
        TestCase(
            query=QueryOption(parameter="t", level_type="pl", level=[850, 925, 1000]),
        ),
        TestCase(
            query=QueryOption(parameter="gh", level_type="isobaricInhPa", level=[850, 925, 1000]),
        )
    ]
    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert field is not None
        assert np.array_equal(
            np.sort(field.coords[test_case.query.level_type].values),
            np.sort(test_case.query.level).astype(float)
        )


@pytest.fixture
def pl_levels():
    return [
        1000,
        975,
        950,
        925,
        900,
        850,
        800,
        750,
        700,
        650,
        600,
        550,
        500,
        450,
        400,
        350,
        300,
        275,
        250,
        225,
        200,
        175,
        150,
        125,
        100,
        70,
        50,
        30,
        20,
        10,
        7,
        5,
        4,
        3,
        2,
        1.5,
        1,
        0.5,
        0.2,
        0.1
    ]


def test_all_levels(file_path, pl_levels):
    test_cases = [
        TestCase(
            query=QueryOption(
                parameter="t",
                level_type="pl",
                level="all"
            ),
            expected_level_name="pl",
            expected_level=pl_levels
        ),
    ]
    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert field is not None
        assert np.array_equal(
            np.sort(field.coords[test_case.expected_level_name].values),
            np.sort(test_case.expected_level).astype(float)
        )


def test_none_level(file_path):
    test_cases = [
        TestCase(
            query=QueryOption(parameter="t", level_type="pl", level=None),
            expected_level_name="pl",
            expected_level=1000
        ),
        TestCase(
            query=QueryOption(parameter="vwsh", level_type="heightAboveGroundLayer", level=None),
            expected_level_name="heightAboveGroundLayer",
            expected_level=1000
        )
    ]
    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.coords[test_case.expected_level_name].values == test_case.expected_level



import pytest
import pandas as pd
from pathlib import Path

from reki.data_finder import find_local_file


def fill_path(file_path: str, start_time: pd.Timestamp, forecast_time: pd.Timedelta):
    start_time_label = start_time.strftime('%Y%m%d%H')
    forecast_hour_label = f"{int(forecast_time/pd.Timedelta(hours=1)):03d}"
    return file_path.format(
        start_time_label=start_time_label,
        forecast_hour_label=forecast_hour_label,
    )


@pytest.mark.parametrize(
    "grib2_type,grib2_path",
    [
        ("orig", "/g3/COMMONDATA/OPER/CEMC/GFS_GMF/Prod-grib/{start_time_label}/ORIG/"
                 "gmf.gra.{start_time_label}{forecast_hour_label}.grb2"),
        ("ne", "/g3/COMMONDATA/OPER/CEMC/GFS_GMF/Prod-grib/{start_time_label}/CMACAST/"
                 "ne_gmf.gra.{start_time_label}{forecast_hour_label}.grb2"),
        ("modelvar", "/g3/COMMONDATA/OPER/CEMC/GFS_GMF/Prod-grib/{start_time_label}/MODELVAR/"
                 "modelvar{start_time_label}{forecast_hour_label}.grb2"),
    ]
)
def test_grib2(cma_gfs_system_name, last_two_day, forecast_time_24h, grib2_type, grib2_path):
    data_type = f"{cma_gfs_system_name}/grib2/{grib2_type}"
    start_time = last_two_day
    forecast_time = forecast_time_24h
    expected_data_path = Path(fill_path(file_path=grib2_path, start_time=start_time, forecast_time=forecast_time))

    data_path = find_local_file(
        data_type=data_type,
        start_time=start_time,
        forecast_time=forecast_time,
    )

    assert data_path is not None
    assert data_path == expected_data_path



import pytest
from typing import Dict, Union, Optional
from dataclasses import dataclass, asdict

from reki.format.grib.eccodes import load_field_from_file


@dataclass
class QueryOption:
    parameter: Union[str, Dict] = None
    level_type: str = None
    level: float = None
    field_name: Optional[str] = None


@dataclass
class TestCase:
    query: QueryOption
    expected_field_name: str


def test_parameter_string(file_path):
    test_cases = [
        TestCase(
            query=QueryOption(parameter="t", level_type="pl", level=850, field_name=None),
            expected_field_name="t"
        ),
        TestCase(
            query=QueryOption(parameter="t", level_type="pl", level=850, field_name="other_field_name"),
            expected_field_name="other_field_name"
        ),
        TestCase(
            query=QueryOption(parameter="TMP", level_type="pl", level=850, field_name=None),
            expected_field_name="TMP"
        ),
        TestCase(
            query=QueryOption(parameter="TMP", level_type="pl", level=850, field_name="other_field_name"),
            expected_field_name="other_field_name"
        )
    ]

    for test_case in test_cases:
        f = load_field_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert f is not None
        assert f.name == test_case.expected_field_name


def test_parameter_cemc_param_db(file_path):
    test_cases = [
        TestCase(
            query=QueryOption(parameter="btv", field_name=None),
            expected_field_name="btv"
        ),
        TestCase(
            query=QueryOption(parameter="zs", field_name=None),
            expected_field_name="zs",
        ),
    ]

    for test_case in test_cases:
        f = load_field_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert f is not None
        assert f.name == test_case.expected_field_name


def test_parameter_dict(file_path):
    test_cases = [
        TestCase(
            query=QueryOption(
                parameter=dict(discipline=0, parameterCategory=2, parameterNumber=224),
                level_type="pl",
                level=850,
                field_name=None,
            ),
            expected_field_name="0_2_224"
        ),
        TestCase(
            query=QueryOption(
                parameter=dict(discipline=0, parameterCategory=2, parameterNumber=224),
                level_type="pl",
                level=850,
                field_name="other_field_name"
            ),
            expected_field_name="other_field_name"
        ),
        TestCase(
            query=QueryOption(
                parameter=dict(discipline=0, parameterCategory=2, parameterNumber=227),
                level_type="sfc",
                field_name=None
            ),
            expected_field_name="0_2_227"
        ),
        TestCase(
            query=QueryOption(
                parameter=dict(discipline=0, parameterCategory=2, parameterNumber=227),
                level_type="sfc",
                field_name="other_field_name"
            ),
            expected_field_name="other_field_name"
        ),
        TestCase(
            query=QueryOption(
                parameter=dict(discipline=0, parameterCategory=5, parameterNumber=4),
                level_type="sfc",
                field_name=None,
            ),
            expected_field_name="ulwrf"
        ),
        TestCase(
            query=QueryOption(
                parameter=dict(discipline=0, parameterCategory=5, parameterNumber=4),
                level_type="sfc",
                field_name="other_field_name",
            ),
            expected_field_name="other_field_name"
        )
    ]

    for test_case in test_cases:
        f = load_field_from_file(
            file_path,
            **asdict(test_case.query),
        )
        assert f is not None
        assert f.name == test_case.expected_field_name



from .message import (
    load_message_from_file,
    load_messages_from_file,
)

from .field import (
    load_field_from_file,
    load_field_from_files,
)

from .bytes import (
    load_bytes_from_file,
    create_message_from_bytes,
    create_messages_from_bytes,
)

from ._xarray import create_data_array_from_message



from typing import Dict, Union, Optional
from pathlib import Path

import xarray as xr

from reki.format.grib.common import fix_level_type

from ._util import (
    _fill_parameter,
    _fill_level,
    _fill_level_type,
    _fill_level_value,
    _fill_index_path,
)
from reki._util import _load_first_variable


def load_field_from_file(
        file_path: Union[str, Path],
        parameter: Union[str, Dict],
        level_type: Union[str, Dict],
        level: Union[int, float] = None,
        with_index: Union[str, bool] = False,
) -> Optional[xr.DataArray]:
    """
    Load **one** field from GRIB2 file using `ecmwf/cfgrib <https://github.com/ecmwf/cfgrib>`_.

    This function loads the first data fitting searching conditions.

    Parameters
    ----------
    file_path: str or Path
        GRIB2 file path
    parameter: str or Dict
        parameter identifier. support two types:

        * str: parameter name, see shortName key using grib_ls of ecCodes.
        * Dict: parameter keys, including:

            * discipline
            * parameterCategory
            * parameterNumber

    level_type: str or Dict
        level type, see typeOfLevel key using grib_ls of ecCodes.
    level: int or float or None
        level value. If none, all levels will be loaded.
    with_index: str or bool
        use index file generated by cfgrib.
        if False, index file will not be used.
        if True, cfgrib will generate index file.

    Returns
    -------
    xr.DataArray or None:
        `xr.DataArray` if found one data, or None if not.

    Examples
    --------
    Read 850hPa temperature from a GRAEPS GFS grib2 file.
    >>> load_field_from_file(
   ...     file_path="/g1/COMMONDATA/OPER/NWPC/GRAPES_GFS_GMF/Prod-grib/2020031721/ORIG/gmf.gra.2020031800105.grb2",
   ...     parameter="t",
   ...     level_type="isobaricInhPa",
   ...     level=850,
   ... )
    <xarray.DataArray 't' (latitude: 720, longitude: 1440)>
    [1036800 values with dtype=float32]
    Coordinates:
        time           datetime64[ns] ...
        step           timedelta64[ns] ...
        isobaricInhPa  int64 ...
      * latitude       (latitude) float64 89.88 89.62 89.38 ... -89.38 -89.62 -89.88
      * longitude      (longitude) float64 0.0 0.25 0.5 0.75 ... 359.2 359.5 359.8
        valid_time     datetime64[ns] ...
    Attributes:
        GRIB_paramId:                             130
        GRIB_shortName:                           t
        GRIB_units:                               K
        GRIB_name:                                Temperature
        GRIB_cfName:                              air_temperature
        GRIB_cfVarName:                           t
        GRIB_dataType:                            fc
        GRIB_missingValue:                        9999
        GRIB_numberOfPoints:                      1036800
        GRIB_typeOfLevel:                         isobaricInhPa
        GRIB_NV:                                  0
        GRIB_stepUnits:                           1
        GRIB_stepType:                            instant
        GRIB_gridType:                            regular_ll
        GRIB_gridDefinitionDescription:           Latitude/longitude
        GRIB_Nx:                                  1440
        GRIB_iDirectionIncrementInDegrees:        0.25
        GRIB_iScansNegatively:                    0
        GRIB_longitudeOfFirstGridPointInDegrees:  0.0
        GRIB_longitudeOfLastGridPointInDegrees:   359.75
        GRIB_Ny:                                  720
        GRIB_jDirectionIncrementInDegrees:        0.25
        GRIB_jPointsAreConsecutive:               0
        GRIB_jScansPositively:                    0
        GRIB_latitudeOfFirstGridPointInDegrees:   89.875
        GRIB_latitudeOfLastGridPointInDegrees:    -89.875
        long_name:                                Temperature
        units:                                    K
        standard_name:                            air_temperature

    Load a filed without shortName.
    >>> load_field_from_file(
    ...     file_path="/g1/COMMONDATA/OPER/NWPC/GRAPES_GFS_GMF/Prod-grib/2020031721/ORIG/gmf.gra.2020031800105.grb2",
    ...     parameter={
    ...         "discipline": 0,
    ...         "parameterCategory": 2,
    ...         "parameterNumber": 225,
    ...     },
    ...     level_type="isobaricInhPa",
    ...     level=850,
    ... )
    <xarray.DataArray 'paramId_0' (latitude: 720, longitude: 1440)>
    [1036800 values with dtype=float32]
    Coordinates:
        time           datetime64[ns] ...
        step           timedelta64[ns] ...
        isobaricInhPa  int64 ...
      * latitude       (latitude) float64 89.88 89.62 89.38 ... -89.38 -89.62 -89.88
      * longitude      (longitude) float64 0.0 0.25 0.5 0.75 ... 359.2 359.5 359.8
        valid_time     datetime64[ns] ...
    Attributes:
        GRIB_paramId:                             0
        GRIB_dataType:                            fc
        GRIB_missingValue:                        9999
        GRIB_numberOfPoints:                      1036800
        GRIB_typeOfLevel:                         isobaricInhPa
        GRIB_NV:                                  0
        GRIB_stepUnits:                           1
        GRIB_stepType:                            instant
        GRIB_gridType:                            regular_ll
        GRIB_gridDefinitionDescription:           Latitude/longitude
        GRIB_Nx:                                  1440
        GRIB_iDirectionIncrementInDegrees:        0.25
        GRIB_iScansNegatively:                    0
        GRIB_longitudeOfFirstGridPointInDegrees:  0.0
        GRIB_longitudeOfLastGridPointInDegrees:   359.75
        GRIB_Ny:                                  720
        GRIB_jDirectionIncrementInDegrees:        0.25
        GRIB_jPointsAreConsecutive:               0
        GRIB_jScansPositively:                    0
        GRIB_latitudeOfFirstGridPointInDegrees:   89.875
        GRIB_latitudeOfLastGridPointInDegrees:    -89.875
        GRIB_discipline:                          0
        GRIB_parameterCategory:                   2
        GRIB_parameterNumber:                     225
        long_name:                                original GRIB paramId: 0
        units:                                    1


    """
    data_set = load_fields_from_file(
        file_path=file_path,
        parameter=parameter,
        level_type=level_type,
        level=level,
        with_index=with_index
    )

    if data_set is None:
        return None

    return _load_first_variable(data_set)


def load_fields_from_file(
        file_path: Union[str, Path],
        parameter: Union[str, Dict] = None,
        level_type: str = None,
        level: Union[int, float] = None,
        with_index: Union[str, bool] = False,
) -> Optional[xr.Dataset]:
    """
    Load fields from GRIB2 file using `ecmwf/cfgrib <https://github.com/ecmwf/cfgrib>`_.

    Parameters
    ----------
    file_path: str or Path
        GRIB2 data file path
    parameter: str or Dict
        see `load_message_from_file`
    level_type: str
        see `load_message_from_file`
    level: int or float or None
        see `load_message_from_file`
    with_index: str or bool
        see `load_message_from_file`

    Returns
    -------
    xr.Dataset or None:
        `xr.Dataset` if found, or None if not.

    Examples
    --------
    >>> load_fields_from_file(
    ...     file_path="/g1/COMMONDATA/OPER/NWPC/GRAPES_GFS_GMF/Prod-grib/2020031721/ORIG/gmf.gra.2020031800105.grb2",
    ...     parameter="t",
    ...     level_type="isobaricInhPa",
    ... )
    <xarray.Dataset>
    Dimensions:        (isobaricInhPa: 36, latitude: 720, longitude: 1440)
    Coordinates:
        time           datetime64[ns] ...
        step           timedelta64[ns] ...
      * isobaricInhPa  (isobaricInhPa) int64 1000 975 950 925 900 850 ... 5 4 3 2 1
      * latitude       (latitude) float64 89.88 89.62 89.38 ... -89.38 -89.62 -89.88
      * longitude      (longitude) float64 0.0 0.25 0.5 0.75 ... 359.2 359.5 359.8
        valid_time     datetime64[ns] ...
    Data variables:
        t              (isobaricInhPa, latitude, longitude) float32 ...
    Attributes:
        GRIB_edition:            2
        GRIB_centre:             babj
        GRIB_centreDescription:  Beijing
        GRIB_subCentre:          0
        Conventions:             CF-1.7
        institution:             Beijing
        history:                 2020-03-20T08:15:39 GRIB to CDM+CF via cfgrib-0....

    """
    filter_by_keys = {}
    read_keys = []

    if parameter is not None:
        _fill_parameter(parameter, filter_by_keys, read_keys)

    if level_type is not None:
        level_type = fix_level_type(level_type, engine="cfgrib")
        _fill_level_type(level_type, filter_by_keys, read_keys)

    if level is not None and level != "all":
        _fill_level_value(level, filter_by_keys, read_keys)

    backend_kwargs = {
        "filter_by_keys": filter_by_keys
    }
    if len(read_keys) > 0:
        backend_kwargs["read_keys"] = read_keys

    _fill_index_path(with_index, backend_kwargs)

    data_set = xr.open_dataset(
        file_path,
        engine="cfgrib",
        backend_kwargs=backend_kwargs
    )

    return data_set



from dataclasses import dataclass, asdict
from typing import Union, Dict, Optional, List

import eccodes

from reki.format.grib.eccodes import load_message_from_file


@dataclass
class QueryOption:
    parameter: Union[str, dict]
    level_type: Union[str, Dict]
    level: Optional[Union[float, Dict, List]]


@dataclass
class TestCase:
    query: QueryOption
    expected_keys: Optional[Dict] = None


def test_short_name(file_path):
    test_cases = [
        TestCase(query=QueryOption(parameter="t", level_type="pl", level=850))
    ]
    for test_case in test_cases:
        message = load_message_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert message is not None
        eccodes.codes_release(message)


def test_numbers(file_path):
    """
    雷达组合反射率
    """
    test_cases = [
        TestCase(query=QueryOption(parameter={
            "discipline": 0,
            "parameterCategory": 16,
            "parameterNumber": 225,
        }, level_type="pl", level=850))
    ]
    for test_case in test_cases:
        message = load_message_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert message is not None
        eccodes.codes_release(message)


def test_embedded_short_name(file_path):
    test_cases = [
        TestCase(query=QueryOption("DEPR", "pl", 850)),
        TestCase(query=QueryOption({
            "discipline": 0,
            "parameterCategory": 0,
            "parameterNumber": 7,
        }, "pl", 850)),
    ]

    messages = []
    for test_case in test_cases:
        message = load_message_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert message is not None
        messages.append(message)

    assert eccodes.codes_get(messages[0], "count") == eccodes.codes_get(messages[1], "count")

    for message in messages:
        eccodes.codes_release(message)



import typing
from typing import Union, List, Dict, Optional
from pathlib import Path

import eccodes
# from tqdm import tqdm

from ._level import _fix_level
from ._util import (
    _check_parameter,
    _check_level_type,
    _check_level_value,
    _check_keys,
    _check_message,
)

from reki.format.grib.common._parameter import convert_parameter


def load_message_from_file(
        file_path: Union[str, Path],
        parameter: Union[str, Dict] = None,
        level_type: Union[str, Dict] = None,
        level: Union[int, float, Dict] = None,
        count: int = None,
        look_parameter: bool = True,
        **kwargs,
) -> Optional[int]:
    """
    Load the **first** message from GRIB 2 file using eccodes-python library.

    Returned message is a copied one of original message and file is closed before return.
    And the returned message should be released by user using `eccodes.codes_release()`.

    Parameters
    ----------
    file_path
        GRIB 2 file path.
    parameter
        short name of the field, or a dict of GRIB keys:

        * discipline
        * parameterCategory
        * parameterNumber

    level_type
        level type.
    level
        level value.
    kwargs
        other grib key used to filter.
    count
        grib message index in grib file, starting with 1.

        This option will make all others options ignored.

    Returns
    -------
    Union[int]
        GRIB handler (int) if found or None if not found.

    Examples
    --------
    Load 850hPa temperature from GRAPES GFS and get values from GRIB message.

    >>> t = load_message_from_file(
    ...     file_path="/g1/COMMONDATA/OPER/NWPC/GRAPES_GFS_GMF/Prod-grib/2020031721/ORIG/gmf.gra.2020031800105.grb2",
    ...     parameter="t",
    ...     level_type="isobaricInhPa",
    ...     level=850,
    ... )
    >>> data = eccodes.codes_get_double_array(t, "values")
    >>> data = data.reshape([720, 1440])
    >>> data
    array([[249.19234375, 249.16234375, 249.16234375, ..., 249.15234375,
        249.19234375, 249.14234375],
       [249.45234375, 249.45234375, 249.42234375, ..., 249.45234375,
        249.44234375, 249.44234375],
       [249.69234375, 249.68234375, 249.68234375, ..., 249.70234375,
        249.67234375, 249.68234375],
       ...,
       [235.33234375, 235.45234375, 235.62234375, ..., 235.47234375,
        235.63234375, 235.48234375],
       [235.78234375, 235.91234375, 235.64234375, ..., 235.80234375,
        235.72234375, 235.82234375],
       [235.66234375, 235.86234375, 235.82234375, ..., 235.85234375,
        235.68234375, 235.70234375]])
    >>> eccodes.codes_release(t)

    """
    if count is not None:
        return _load_message_from_file_by_count(file_path, count)

    fixed_level_type, _ = _fix_level(level_type, None)

    if look_parameter:
        parameter = convert_parameter(parameter)

    with open(file_path, "rb") as f:
        while True:
            message_id = eccodes.codes_grib_new_from_file(f)
            if message_id is None:
                return None
            if not _check_message(message_id, parameter, fixed_level_type, level, **kwargs):
                eccodes.codes_release(message_id)
                continue
            return message_id

            # # clone message
            # new_message_id = eccodes.codes_clone(message_id)
            # eccodes.codes_release(message_id)
            # return new_message_id
        return None


def load_messages_from_file(
        file_path: Union[str, Path],
        parameter: Union[str, Dict],
        level_type: Union[str, Dict, List] = None,
        level: Union[int, float, List, Dict] = None,
        **kwargs,
) -> Optional[List]:
    """
    Load multiply messages from file.

    This function will scan all messages in GRIB 2 file and return all messages which fit conditions.

    Parameters
    ----------
    file_path: str or Path
    parameter: str or typing.Dict
        see ``load_message_from_file``, required option.
    level_type: str or typing.List or None
        level type.
        - string, same as ``load_message_from_file``
        - typing.List, level type should be in the list.
        - None, don't check level type.
    level: int or float or typing.Dict or typing.List or None
        level value.
        - string, same as ``load_message_from_file``
        - typing.Dict, same as ``load_message_from_file``
        - typing.List, level value should be in the list.
        - None, don't check level value. For example, load all messages of some typeOfLevel.
    kwargs: dict
        other grib key used to filter.

    Returns
    -------
    typing.List or None:
        a list of message number or None if no message is found.
    """
    fixed_level_type, _ = _fix_level(level_type, None)

    parameter = convert_parameter(parameter)

    messages = []

    # print("count...")
    # with open(file_path, "rb") as f:
    #     total_count = eccodes.codes_count_in_file(f)
    #     print(total_count)
    # print("count..done")

    with open(file_path, "rb") as f:
        # pbar = tqdm(total=total_count)
        while True:
            message_id = eccodes.codes_grib_new_from_file(f)
            if message_id is None:
                break
            # pbar.update(1)
            if not _check_parameter(message_id, parameter):
                eccodes.codes_release(message_id)
                continue
            if not _check_level_type(message_id, fixed_level_type):
                eccodes.codes_release(message_id)
                continue
            if not _check_level_value(message_id, level):
                eccodes.codes_release(message_id)
                continue
            if not _check_keys(message_id, **kwargs):
                eccodes.codes_release(message_id)
                continue

            # clone message
            new_message_id = eccodes.codes_clone(message_id)
            eccodes.codes_release(message_id)
            messages.append(new_message_id)
        # pbar.close()
        if len(messages) == 0:
            return None
        return messages


def _load_message_from_file_by_count(file_path, count):
    current_index = 0
    with open(file_path, "rb") as f:
        while True:
            message_id = eccodes.codes_grib_new_from_file(f)
            if message_id is None:
                return None
            current_index += 1
            if current_index == count:
                return message_id
            else:
                eccodes.codes_release(message_id)



from dataclasses import dataclass, asdict
from typing import Dict, List, Union, Optional

import numpy as np
import pytest

from reki.format.grads import load_field_from_file


@dataclass
class QueryOption:
    parameter: str
    level_type: Optional[str]
    level: Optional[Union[int, float, List]]


@dataclass
class TestCase:
    query: QueryOption
    expected_level_type: str
    expected_level: Union[int, float, List]


def test_number(file_path, forecast_time):
    test_cases = [
        TestCase(query=QueryOption("t", "pl", 850), expected_level_type="pl", expected_level=850),
        TestCase(query=QueryOption("h", "pl", 500.0), expected_level_type="pl", expected_level=500.0),
        TestCase(query=QueryOption("Qv", "index", 0), expected_level_type="level", expected_level=1000),
        TestCase(query=QueryOption("u", "index", 1), expected_level_type="level", expected_level=975),
        TestCase(query=QueryOption("tslb", "index", 2), expected_level_type="level", expected_level=950),
        TestCase(query=QueryOption("q2m", "single", 0), expected_level_type="level", expected_level=0),
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            forecast_time=forecast_time,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.name == test_case.query.parameter
        assert field.coords[test_case.expected_level_type] == test_case.expected_level


def test_list(file_path, modelvar_file_path, forecast_time):
    test_cases = [
        TestCase(
            query=QueryOption("t", "pl", [1000, 850, 500]),
            expected_level_type="pl", expected_level=[1000, 850, 500]
        ),
        TestCase(
            query=QueryOption("tslb", "index", [0, 1, 2, 3]),
            expected_level_type="level", expected_level=[1000, 975, 950, 925]
        )
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            forecast_time=forecast_time,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.name == test_case.query.parameter
        assert np.array_equal(
            np.sort(field.coords[test_case.expected_level_type].values),
            np.sort(test_case.expected_level)
        )

    test_cases = [
        TestCase(
            query=QueryOption("pi", "ml", [1, 20, 30, 50]),
            expected_level_type="ml", expected_level=[1, 20, 30, 50]
        ),
        TestCase(
            query=QueryOption("Qc", "index", [0, 1, 2, 3]),
            expected_level_type="level", expected_level=[1, 2, 3, 4]
        )
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            modelvar_file_path,
            forecast_time=forecast_time,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.name == test_case.query.parameter
        assert np.array_equal(
            np.sort(field.coords[test_case.expected_level_type].values),
            np.sort(test_case.expected_level)
        )


@pytest.fixture
def pl_levels():
    return [
        1000.0,
        975.0,
        950.0,
        925.0,
        900.0,
        850.0,
        800.0,
        750.0,
        700.0,
        650.0,
        600.0,
        550.0,
        500.0,
        450.0,
        400.0,
        350.0,
        300.0,
        250.0,
        200.0,
        150.0,
        100.0,
        70.0,
        50.0,
        30.0,
        20.0,
        10.0,
    ]


def test_none(file_path, pl_levels, forecast_time):
    test_cases = [
        TestCase(query=QueryOption("u10m", "single", None), expected_level_type="level", expected_level=0),
        TestCase(query=QueryOption("tmn", None, None), expected_level_type="level", expected_level=0)
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            forecast_time=forecast_time,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.name == test_case.query.parameter
        assert field.coords[test_case.expected_level_type] == test_case.expected_level

    test_cases = [
        TestCase(
            query=QueryOption("t", "pl", None),
            expected_level_type="pl", expected_level=pl_levels
        ),
        TestCase(
            query=QueryOption("tslb", None, None),
            expected_level_type="level", expected_level=[1000, 975, 950, 925]
        ),
    ]

    for test_case in test_cases:
        field = load_field_from_file(
            file_path,
            forecast_time=forecast_time,
            **asdict(test_case.query)
        )
        assert field is not None
        assert field.name == test_case.query.parameter
        assert np.array_equal(
            np.sort(field.coords[test_case.expected_level_type].values),
            np.sort(test_case.expected_level)
        )



from typing import BinaryIO

import numpy as np

from .grads_ctl import GradsCtl


class GradsRecordHandler(object):
    """
    Load record data from binary file.
    """
    def __init__(self, grads_ctl: GradsCtl, record_index: int, offset: int, var_index: int = -1, level_index: int = -1):
        self.grads_ctl = grads_ctl
        self.record_index = record_index
        self.record_info = grads_ctl.record[record_index]
        self.offset = offset

        self.var_index = var_index
        self.level_index = level_index

        self.data = None

    def load_data(self, data_file: BinaryIO) -> np.ndarray:
        if 'sequential' in self.grads_ctl.options:
            self.offset += 4
        x_count = self.grads_ctl.xdef['count']
        y_count = self.grads_ctl.ydef['count']

        if self.grads_ctl.data_endian == 'big':
            data_format = '>f'
        elif self.grads_ctl.data_endian == 'little':
            data_format = '<f'
        else:
            # print("Data endian is not found. Use local endian to unpack values.")
            data_format = 'f'

        # load data from file
        data_file.seek(self.offset)

        var_yrev = self.grads_ctl.yrev

        values = np.fromfile(data_file, dtype=np.dtype(data_format), count=x_count * y_count)
        values = values.reshape((y_count, x_count))

        if var_yrev:
            values = np.flip(values, 0)

        self.data = values
        return values



import sys
import re
from pathlib import Path
import logging
from typing import Optional, Union

import pandas as pd


logger = logging.getLogger(__name__)


class GradsCtl(object):
    def __init__(self):
        self.dset = None  # data file path
        self.dset_template = False

        self.title = ''
        self.options = list()
        self.data_endian = 'little'
        self.local_endian = sys.byteorder
        self.yrev = 0
        self.undef = None

        self.start_time = None
        self.forecast_time = None

        # dimension
        self.xdef = None
        self.ydef = None
        self.zdef = None
        self.tdef = None

        self.vars = []
        self.record = []

    def get_data_file_path(self, record):
        return GradsCtlParser.get_data_file_path(self, record)


class GradsCtlParser(object):
    def __init__(self, grads_ctl: Optional[GradsCtl] = None):
        self.ctl_file_path = None

        if grads_ctl is None:
            grads_ctl = GradsCtl()
        self.grads_ctl = grads_ctl

        self.ctl_file_lines = list()
        self.cur_no = -1

        self.parser_mapper = {
            'ctl_file_name': self._parse_ctl_file_name,
            'dset': self._parse_dset,
            'options': self._parse_options,
            'title': self._parse_title,
            'undef': self._parse_undef,
            'xdef': self._parse_dimension,
            'ydef': self._parse_dimension,
            'zdef': self._parse_dimension,
            'tdef': self._parse_tdef,
            'vars': self._parse_vars,
        }

    def set_ctl_file_path(self, ctl_file_path: Union[Path, str]):
        self.ctl_file_path = Path(ctl_file_path)
        with open(ctl_file_path) as f:
            lines = f.readlines()
            self.ctl_file_lines = [l.strip() for l in lines]
            self.cur_no = 0

    def parse(self, ctl_file_path: Union[Path, str]):
        self.set_ctl_file_path(ctl_file_path)
        total_lines = len(self.ctl_file_lines)
        while self.cur_no < total_lines:
            cur_line = self.ctl_file_lines[self.cur_no]
            first_word = cur_line[0:cur_line.find(' ')]
            if first_word.lower() in self.parser_mapper:
                self.parser_mapper[first_word]()
            self.cur_no += 1

        self._parse_ctl_file_name()

        return self.grads_ctl

    def _parse_ctl_file_name(self):
        ctl_file_name = Path(self.ctl_file_path).name

        if (
                self.grads_ctl.start_time is None
                and self.grads_ctl.forecast_time is None
        ):
            logger.debug("guess start time and forecast time")

            if ctl_file_name.startswith("post.ctl_") or ctl_file_name.startswith("model.ctl_"):
                file_name = ctl_file_name[ctl_file_name.index("_")+1:]
                # check for GRAPES MESO:
                #   post.ctl_201408111202900
                if re.match(r"[0-9]{15}", file_name):
                    self.grads_ctl.start_time = pd.to_datetime(file_name[:10], format="%Y%m%d%H")
                    self.grads_ctl.forecast_time = pd.Timedelta(hours=int(file_name[11:14]))
                # check for GRAPES GFS:
                #   post.ctl_2014081112_001
                elif re.match(r"[0-9]{10}_[0-9]{3}", file_name):
                    self.grads_ctl.start_time = pd.to_datetime(file_name[:10], format="%Y%m%d%H")
                    self.grads_ctl.forecast_time = pd.Timedelta(hours=int(file_name[12:]))
                else:
                    logger.warning("We can't recognize ctl file name. ")

    def _parse_dset(self):
        """
        parse data file path:

            dset ^postvar2021080200_024
            dset ^postvar2021092700%f3%n2

        """
        cur_line = self.ctl_file_lines[self.cur_no]
        file_path = cur_line[4:].strip()
        if "%" in file_path:
            self.grads_ctl.dset_template = True
        if file_path[0] == '^':
            file_dir = Path(self.ctl_file_path).parent
            file_path = Path(file_dir, file_path[1:])

        self.grads_ctl.dset = file_path

    def _parse_options(self):
        cur_line = self.ctl_file_lines[self.cur_no]
        options = cur_line[7:].strip().split(' ')
        self.grads_ctl.options.extend(options)
        for an_option in options:
            if an_option == 'big_endian':
                self.grads_ctl.data_endian = 'big'
            elif an_option == 'little_endian':
                self.grads_ctl.data_endian = 'little'
            elif an_option == 'yrev':
                self.grads_ctl.yrev = True

    def _parse_title(self):
        cur_line = self.ctl_file_lines[self.cur_no]
        title = cur_line[5:].strip()
        self.grads_ctl.title = title

    def _parse_undef(self):
        cur_line = self.ctl_file_lines[self.cur_no]
        undef = cur_line[5:].strip()
        self.grads_ctl.undef = float(undef)

    def _parse_dimension(self):
        """
        parser for keywords xdef, ydef and zdef
        """
        cur_line = self.ctl_file_lines[self.cur_no].lower()
        tokens = cur_line.split()
        dim_name = tokens[0]  # xdef, ydef, zdef
        dimension_type = tokens[2]

        dimension_parser_map = {
            'linear': self._parse_linear_dimension,
            'levels': self._parse_levels_dimension
        }

        if dimension_type in dimension_parser_map:
            dimension_parser_map[dimension_type](dim_name, tokens)
        else:
            raise NotImplemented(f'dimension_type is not supported: {dimension_type}')

    def _parse_linear_dimension(self, dim_name, tokens):
        """
        Parse linear dimension

            xdef 1440 linear    0.0000    0.2500

        """
        if len(tokens) < 4:
            raise Exception("%s parser error" % dim_name)
        count = int(tokens[1])
        start = float(tokens[3])
        step = float(tokens[4])
        levels = [start + step * n for n in range(count)]
        setattr(self.grads_ctl, dim_name, {
            'type': 'linear',
            'count': count,
            'start': start,
            'step': step,
            'values': levels
        })

    def _parse_levels_dimension(self, dim_name, tokens):
        """
        Parse levels dimension

            zdef   27 levels
                 1000.000
                 925.0000
                 850.0000
                 700.0000
                 ...

            zdef 22 levels
            1000 925 850 700 600 500 400 300 250 200 150 100 70 50 30 20 10 7 5 3 2 1

        """
        levels = list()
        count = int(tokens[1])
        if len(tokens) > 2:
            levels = [float(l) for l in tokens[3:]]
        i = len(levels)
        while i < count:
            self.cur_no += 1
            cur_line = self.ctl_file_lines[self.cur_no]
            current_line_tokens = cur_line.split()
            current_levels = [float(l) for l in current_line_tokens]
            levels.extend(current_levels)
            i += len(current_line_tokens)

        setattr(self.grads_ctl, dim_name, {
            'type': 'levels',
            'count': count,
            'values': levels
        })

    def _parse_tdef(self):
        """
        Parse time dimension

            tdef    1 linear 00z03AUG2021   360mn

        """
        cur_line = self.ctl_file_lines[self.cur_no]
        parts = cur_line.strip().split()
        assert parts[2] == "linear"
        assert len(parts) == 5

        count = int(parts[1])
        start_string = parts[3]
        increment_string = parts[4]

        start_date = GradsCtlParser._parse_start_time(start_string)
        time_step = GradsCtlParser._parse_increment_time(increment_string)

        values = [start_date + time_step * i for i in range(count)]

        self.grads_ctl.tdef = {
            'type': 'linear',
            'count': count,
            'start': start_date,
            'step': time_step,
            'values': values
        }

    @classmethod
    def _parse_start_time(cls, start_string) -> pd.Timestamp:
        """
        parse start time
        format:
            hh:mmZddmmmyyyy
        where:
            hh	=	hour (two digit integer)
            mm	=	minute (two digit integer)
            dd	=	day (one or two digit integer)
            mmm	=	3-character month
            yyyy	=	year (may be a two or four digit integer;
                              2 digits implies a year between 1950 and 2049)
        """
        start_date = pd.Timestamp.now()
        if start_string[3] == ':':
            raise NotImplemented('Not supported time with hh')
        elif len(start_string) == 12:
            start_date = pd.to_datetime(start_string.lower(), format='%Hz%d%b%Y')
        else:
            raise NotImplemented(f'start time not supported: {start_string}')
        return start_date

    @classmethod
    def _parse_increment_time(cls, increment_string) -> pd.Timedelta:
        """
        parse increment time
        format:
            vvkk
        where:
            vv	=	an integer number, 1 or 2 digits
            kk	=	mn (minute)
                    hr (hour)
                    dy (day)
                    mo (month)
                    yr (year)
        """
        vv = int(increment_string[:-2])
        kk = increment_string[-2:]

        kk_map = {
            'mn': (lambda v: pd.Timedelta(minutes=v)),
            'hr': (lambda v: pd.Timedelta(hours=v)),
            'dy': (lambda v: pd.Timedelta(days=v))
        }
        if kk in kk_map:
            return kk_map[kk](vv)
        else:
            raise NotImplemented('{kk} is not supported'.format(kk=kk))

    def _parse_vars(self):
        self._parse_variables()
        self._generate_records()

    def _parse_variables(self):
        var_list = list()

        parts = self.ctl_file_lines[self.cur_no].strip().split()
        assert len(parts) == 2
        count = int(parts[1])
        for _ in range(count):
            # parse one var line
            self.cur_no += 1
            cur_line = self.ctl_file_lines[self.cur_no].strip()
            parts = cur_line.split()
            # we currently use old style of var record.
            # TODO: check for new type of record from GrADS v2.0.2

            var_name = parts[0]
            levels = int(parts[1])
            units = parts[2]
            description = " ".join(parts[3:])

            cur_var = {
                'name': var_name,
                'levels': levels,
                'units': units,
                'description': description
            }
            var_list.append(cur_var)

        self.grads_ctl.vars = var_list

    def _generate_records(self):
        record_list = list()
        record_index = 0
        for time_step_index, valid_time in enumerate(self.grads_ctl.tdef["values"]):
            forecast_time = self.grads_ctl.tdef["step"] * time_step_index
            if self.grads_ctl.dset_template:
                record_index = 0
            for a_var_record in self.grads_ctl.vars:
                if a_var_record['levels'] == 0:
                    record_list.append({
                        'name': a_var_record['name'],
                        'level_type': 'single',
                        'level': 0,
                        'level_index': 0,
                        'valid_time': valid_time,
                        'forecast_time': forecast_time,
                        'units': a_var_record['units'],
                        'description': a_var_record['description'],
                        'record_index': record_index
                    })
                    record_index += 1
                else:
                    for level_index in range(0, a_var_record["levels"]):
                        a_level = self.grads_ctl.zdef["values"][level_index]
                        record_list.append({
                            'name': a_var_record['name'],
                            'level_type': 'multi',
                            'level': a_level,
                            'level_index': level_index,
                            'units': a_var_record['units'],
                            'valid_time': valid_time,
                            'forecast_time': forecast_time,
                            'description': a_var_record['description'],
                            'record_index': record_index
                        })
                        record_index += 1

        self.grads_ctl.record = record_list

    @classmethod
    def get_data_file_path(cls, grads_ctl, record) -> Path:
        if not grads_ctl.dset_template:
            return grads_ctl.dset

        tokens = str(grads_ctl.dset).split("%")
        token_mapper = {
            "y2": lambda x: f"{x['valid_time'].year % 100:02d}",
            "y4": lambda x: f"{x['valid_time'].year:04d}",
            "m1": lambda x: f"{x['valid_time'].month}",
            "m2": lambda x: f"{x['valid_time'].month:02d}",
            "d1": lambda x: f"{x['valid_time'].day}",
            "d2": lambda x: f"{x['valid_time'].day:02d}",
            "h1": lambda x: f"{x['valid_time'].hour}",
            "h2": lambda x: f"{x['valid_time'].hour:02d}",
            "h3": lambda x: f"{x['valid_time'].hour:03d}",
            "n2": lambda x: f"{x['forecast_time'].seconds // 60 % 60:02d}",  # TODO: check n2
            "f2": lambda x: f"{int(x['forecast_time'] / pd.Timedelta(hours=1)):02d}",
            "f3": lambda x: f"{int(x['forecast_time'] / pd.Timedelta(hours=1)):03d}",
            "fn2": lambda x: f"{int(x['forecast_time'] / pd.Timedelta(minutes=1)):02d}",
            "fhn": lambda x: f"{int(x['forecast_time'] / pd.Timedelta(hours=1)):02d}{int(x['forecast_time'].total_seconds() % (60*60) / 60):02d}"
        }

        def parse_file_template_token(token: str) -> str:
            for key in token_mapper:
                if token.startswith(key):
                    result = token_mapper[key](record) + token[len(key):]
                    return result
            return token

        parts = tokens[:1]
        for token in tokens[1:]:
            parsed_part = parse_file_template_token(token)
            parts.append(parsed_part)

        return Path("".join(parts))



import pytest

from reki.data_finder import find_local_file
from reki.format.grib.eccodes import load_field_from_file
from reki.operator import extract_point


@pytest.fixture
def cma_gfs_grib2_orig_file_path(last_two_day, forecast_time_24h):
    start_time = last_two_day
    forecast_time = forecast_time_24h
    file_path = find_local_file(
        f"cma_gfs_gmf/grib2/orig",
        start_time=start_time,
        forecast_time=forecast_time,
    )
    return file_path


@pytest.fixture
def t_2m_field(cma_gfs_grib2_orig_file_path):
    file_path = cma_gfs_grib2_orig_file_path
    parameter = "2t"

    field = load_field_from_file(
        file_path=file_path,
        parameter=parameter,
    )
    return field


def test_extract_point(t_2m_field):
    data = t_2m_field
    point = extract_point(
        data=data,
        longitude=116.17,
        latitude=39.56,
    )
    assert point.latitude.values.item() == 39.56
    assert point.longitude.values.item() == 116.17


@pytest.mark.parametrize(
    "engine,scheme",
    [
        ("xarray", "linear"),
        ("xarray", "nearest"),
        ("scipy", "linear"),
        ("scipy", "nearest"),
        ("scipy", "splinef2d"),
        ("scipy", "rect_bivariate_spline"),
    ]
)
def test_extract_point_with_engine_and_scheme(t_2m_field, engine, scheme):
    data = t_2m_field
    point = extract_point(
        data=data,
        longitude=116.17,
        latitude=39.56,
        scheme=scheme,
        engine=engine,
    )
    assert point.latitude.values.item() == 39.56
    assert point.longitude.values.item() == 116.17


def test_extract_point_multi_points(t_2m_field):
    data = t_2m_field
    point = extract_point(
        data=data,
        longitude=[115, 116, 117],
        latitude=[40, 39],
    )
    assert point.shape == (2, 3)
    assert list(point.latitude.values) == [40, 39]
    assert list(point.longitude.values) == [115, 116, 117]


@pytest.mark.parametrize(
    "engine,scheme",
    [
        ("xarray", "linear"),
        ("xarray", "nearest"),
        ("scipy", "linear"),
        ("scipy", "nearest"),
        ("scipy", "splinef2d"),
        ("scipy", "rect_bivariate_spline"),
    ]
)
def test_extract_point_multi_points_with_engine_and_scheme(t_2m_field, engine, scheme):
    data = t_2m_field
    point = extract_point(
        data=data,
        longitude=[115, 116, 117],
        latitude=[40, 39],
        scheme=scheme,
        engine=engine,
    )
    assert point.shape == (2, 3)


@pytest.mark.parametrize(
    "engine,scheme",
    [
        ("xarray", "linear"),
        ("xarray", "nearest"),
        ("scipy", "linear"),
        ("scipy", "nearest"),
        ("scipy", "splinef2d"),
        ("scipy", "rect_bivariate_spline"),
    ]
)
def test_extract_point_multi_points_with_engine_and_scheme_lat(t_2m_field, engine, scheme):
    data = t_2m_field
    point = extract_point(
        data=data,
        longitude=[115, 116, 117],
        latitude=[39, 40],
        scheme=scheme,
        engine=engine,
    )
    assert point.shape == (2, 3)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: transport.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='transport.proto',
  package='transport',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=b'\n\x0ftransport.proto\x12\ttransport\"\x1d\n\x08RawField\x12\x11\n\traw_bytes\x18\x01 \x02(\x0c'
)




_RAWFIELD = _descriptor.Descriptor(
  name='RawField',
  full_name='transport.RawField',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='raw_bytes', full_name='transport.RawField.raw_bytes', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30,
  serialized_end=59,
)

DESCRIPTOR.message_types_by_name['RawField'] = _RAWFIELD
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

RawField = _reflection.GeneratedProtocolMessageType('RawField', (_message.Message,), {
  'DESCRIPTOR' : _RAWFIELD,
  '__module__' : 'transport_pb2'
  # @@protoc_insertion_point(class_scope:transport.RawField)
  })
_sym_db.RegisterMessage(RawField)


# @@protoc_insertion_point(module_scope)



from typing import Union, Dict, Optional

import xarray as xr
import numpy as np
import eccodes

from reki.operator.area import extract_region as extract_region_field
from reki.operator.regrid import interpolate_grid as interpolate_grid_field
from reki.format.grib.eccodes._xarray import create_data_array_from_message
from reki.format.grib.common import MISSING_VALUE


def extract_region(
        message,
        start_longitude: Union[float, int],
        end_longitude: Union[float, int],
        start_latitude: Union[float, int],
        end_latitude: Union[float, int],
        longitude_step: Optional[Union[float, int]] = None,
        latitude_step: Optional[Union[float, int]] = None,
):
    """
    extract region from gridded data array.

    Parameters
    ----------
    message
    start_longitude
    end_longitude
    start_latitude
    end_latitude
    longitude_step
    latitude_step

    Returns
    -------
    """
    missing_value = MISSING_VALUE
    field = create_data_array_from_message(
        message,
        missing_value=missing_value,
        fill_missing_value=np.nan,
    )

    target_field = extract_region_field(
        field,
        start_longitude=start_longitude,
        end_longitude=end_longitude,
        start_latitude=start_latitude,
        end_latitude=end_latitude,
        longitude_step=longitude_step,
        latitude_step=latitude_step
    )

    eccodes.codes_set_double(message, 'longitudeOfFirstGridPointInDegrees', target_field.longitude.values[0])
    eccodes.codes_set_double(message, 'longitudeOfLastGridPointInDegrees', target_field.longitude.values[-1])
    eccodes.codes_set_double(message, 'iDirectionIncrementInDegrees', abs(target_field.longitude.values[0] - target_field.longitude.values[1]))
    # eccodes.codes_set_double(message, 'iDirectionIncrementInDegrees', self.grid.lon_step)
    eccodes.codes_set_long(message, 'Ni', len(target_field.longitude.values))

    eccodes.codes_set_double(message, 'latitudeOfFirstGridPointInDegrees', target_field.latitude.values[0])
    eccodes.codes_set_double(message, 'latitudeOfLastGridPointInDegrees', target_field.latitude.values[-1])
    eccodes.codes_set_double(message, 'jDirectionIncrementInDegrees', abs(target_field.latitude.values[0] - target_field.latitude.values[1]))
    # eccodes.codes_set_double(message, 'jDirectionIncrementInDegrees', self.grid.lat_step)
    eccodes.codes_set_long(message, 'Nj', len(target_field.latitude.values))

    values = target_field.values.flatten()

    np.place(values, np.isnan(values), missing_value)
    # values = np.nan_to_num(values, missing_value)

    eccodes.codes_set(message, 'missingValue', missing_value)
    num_missing = 0
    for i in range(len(values)):
        if values[i] == missing_value:
            num_missing += 1

    if num_missing > 0:
        eccodes.codes_set(message, 'bitmapPresent', 1)

    # close constant field feature.
    # eccodes.codes_set_long(message, "produceLargeConstantFields", 1)

    eccodes.codes_set_double_array(message, "values", values)

    del field
    del target_field

    return message


def interpolate_grid(
        message,
        latitude,
        longitude,
        scheme: str = "linear",
        engine: str = "scipy",
        **kwargs: Dict,
):
    """

    Parameters
    ----------
    message
    latitude
    longitude
        interpolate method.
    scheme
    engine
        interpolate engine, `scipy` or `xarray`
    **kwargs

    Returns
    -------

    """
    target_grid = xr.DataArray(
        coords=[
            ("latitude", latitude),
            ("longitude", longitude)
        ]
    )

    missing_value = MISSING_VALUE
    field = create_data_array_from_message(
        message,
        missing_value=MISSING_VALUE,
        fill_missing_value=np.nan,
    )

    options = kwargs.copy()
    original_fill_value = None
    if "fill_value" in options:
        original_fill_value = options["fill_value"]
        options["fill_value"] = np.nan

    target_field = interpolate_grid_field(
        data=field,
        target=target_grid,
        scheme=scheme,
        engine=engine,
        **options
    )

    eccodes.codes_set_double(message, 'longitudeOfFirstGridPointInDegrees', target_field.longitude.values[0])
    eccodes.codes_set_double(message, 'longitudeOfLastGridPointInDegrees', target_field.longitude.values[-1])
    eccodes.codes_set_double(message, 'iDirectionIncrementInDegrees', abs(target_field.longitude.values[0] - target_field.longitude.values[1]))
    eccodes.codes_set_long(message, 'Ni', len(target_field.longitude.values))

    eccodes.codes_set_double(message, 'latitudeOfFirstGridPointInDegrees', target_field.latitude.values[0])
    eccodes.codes_set_double(message, 'latitudeOfLastGridPointInDegrees', target_field.latitude.values[-1])
    eccodes.codes_set_double(message, 'jDirectionIncrementInDegrees', abs(target_field.latitude.values[0] - target_field.latitude.values[1]))
    eccodes.codes_set_long(message, 'Nj', len(target_field.latitude.values))

    values = target_field.values.flatten()

    if original_fill_value is not None:
        np.place(values, np.isnan(values), missing_value)

    eccodes.codes_set(message, 'missingValue', missing_value)
    num_missing = 0
    for i in range(len(values)):
        if values[i] == missing_value:
            num_missing += 1

    if num_missing > 0:
        eccodes.codes_set(message, 'bitmapPresent', 1)

    values = np.nan_to_num(values, missing_value)
    eccodes.codes_set_values(message, values)

    if num_missing > 0:
        num_data = eccodes.codes_get(message, 'numberOfDataPoints', int)
        assert num_data == len(values)
        assert eccodes.codes_get(message, 'numberOfCodedValues', int) == num_data - num_missing
        assert eccodes.codes_get(message, 'numberOfMissing', int) == num_missing

    del field
    del target_field

    return message



from dataclasses import dataclass, asdict
from typing import Union, Dict, Optional, List

import eccodes

from reki.format.grib.eccodes import load_message_from_file


@dataclass
class QueryOption:
    count: int


@dataclass
class TestCase:
    query: QueryOption
    expected_keys: Dict


def test_count(file_path):
    test_cases = [
        TestCase(query=QueryOption(count=10), expected_keys=dict(count=10)),
        TestCase(query=QueryOption(count=20), expected_keys=dict(count=20)),
    ]

    for test_case in test_cases:
        message = load_message_from_file(
            file_path,
            **asdict(test_case.query)
        )
        assert message is not None
        for key, expected_value in test_case.expected_keys.items():
            assert eccodes.codes_get(message, key, ktype=int) == expected_value

        eccodes.codes_release(message)



"""
Short name table from WGRIB2 project.

Collect these short names from the following systems:

- GRAPES GFS
- GRAPES MESO
- GRAPES GEPS
- GRAPES MEPS
"""

import io

import pandas as pd


_WGRIB2_SHORT_NAME_TABLE_CONTENT = """short_name,discipline,parameterCategory,parameterNumber
ACPCP,0,1,10
ALBDO,0,19,1
APCP,0,1,8
ASNOW,0,1,29
BLI,0,7,1
CAPE,0,7,6
CDCC,0,6,22
CIN,0,7,7
CLWMR,0,1,22
CR,0,16,224
DCAPE,0,7,224
DEPR,0,0,7
DPT,0,0,6
DZDT,0,2,9
EPOT,0,0,3
FRZR,0,1,225
GRLE,0,1,32
GUST,0,2,22
HCDC,0,6,5
HFLUX,2,0,24
HGT,0,3,5
HI,0,1,239
HLCY,0,7,8
HPBL,0,3,18
ICMR,0,1,23
KX,0,7,2
LCDC,0,6,3
LHTFL,0,0,10
MCDC,0,6,4
NCPCP,0,1,9
NLWRF,0,5,5
NSWRF,0,4,9
NSWRFCS,0,4,11
PLI,0,7,0
PRES,0,3,0
PRMSL,0,3,1
PTYPE,0,1,19
PWAT,0,1,3
RELD,0,2,13
RELV,0,2,12
RETOP,0,16,3
RH,0,1,1
RI,0,7,12
RWMR,0,1,24
SHWINX,0,7,13
SKINT,0,0,17
SNMR,0,1,25
SNOD,0,1,11
SOIL_M,2,0,22
SPFH,0,1,0
SX,0,7,5
TCDC,0,6,1
TCIWV,0,1,64
TCOLI,0,1,70
TCOLW,0,1,69
TMAX,0,0,4
TMIN,0,0,5
TMP,0,0,0
TOTALX,0,7,4
TSOIL,2,0,2
TSRWE,0,1,53
UGRD,0,2,2
ULWRF,0,5,4
UPHL,0,7,15
USWRF,0,4,8
VGRD,0,2,3
VIS,0,19,0
VWSH,0,2,25
"""


def _get_wgrib2_short_name_table() -> pd.DataFrame:
    f = io.StringIO(_WGRIB2_SHORT_NAME_TABLE_CONTENT)
    df = pd.read_table(
        f,
        header=0,
        sep=",",
    )
    return df


WGRIB2_SHORT_NAME_TABLE = _get_wgrib2_short_name_table()



import io

import pandas as pd


_CEMC_PARAM_TABLE_CONTENT = """name,unit,description,description_cn,discipline,category,number,typeOfLevel,level,first_level_type,first_level,second_level_type,second_level,stepType,alias
ps,Pa,Surface pressure,地面气压,0.0,3.0,0.0,surface,,1.0,,,,,False
psfc,Pa,Surface pressure,地面气压,0.0,3.0,0.0,surface,,1.0,,,,,True
psl,Pa,Sea level pressure,海平面气压,0.0,3.0,1.0,meanSea,,101.0,,,,,False
pi,,Exner pressure,Exner气压,0.0,3.0,26.0,,,,,,,,False
h,Gpm,Geopotential height,位势高度,0.0,3.0,5.0,,,100.0,,,,,False
zs,Gpm,Terrain ,地形高度,0.0,3.0,5.0,surface,,1.0,,,,,False
t,K,Temperature,温度,0.0,0.0,0.0,,,,,,,,False
t2m,K,Temperature at 2 meters height,2米温度,0.0,0.0,0.0,heightAboveGround,2.0,103.0,2.0,,,,False
ts,K,Surface temperature,地面温度,0.0,0.0,0.0,surface,,1.0,,,,,False
tsfc,K,Surface temperature,地面温度,0.0,0.0,0.0,surface,,1.0,,,,,True
thetaSe,K,Pseudo-equivalent potential temperature,假相当位温,0.0,0.0,3.0,,,,,,,,False
mx2t,K,Maximum of 2 meters high temperature,2米最高温度,0.0,0.0,4.0,heightAboveGround,2.0,103.0,2.0,,,max,False
tmax2m,K,Maximum of 2 meters high temperature,2米最高温度,0.0,0.0,4.0,heightAboveGround,2.0,103.0,2.0,,,max,True
mn2t,K,Minimum of 2 meters high temperature,2米最低温度,0.0,0.0,5.0,heightAboveGround,2.0,103.0,2.0,,,min,False
tmin2m,K,Minimum of 2 meters high temperature,2米最低温度,0.0,0.0,5.0,heightAboveGround,2.0,103.0,2.0,,,min,True
td,K,Dew point,露点温度,0.0,0.0,6.0,,,,,,,,False
td2m,K,2 meters high Dew point,2米露点温度,0.0,0.0,6.0,heightAboveGround,2.0,103.0,2.0,,,,False
ttd,K,Dew point depression,温度露点差,0.0,0.0,7.0,,,,,,,,False
vis,m,Visibility,能见度,0.0,19.0,0.0,surface,,1.0,,,,,False
pli,K,Parcel lifted index(to 500 hpa),气块抬升指数,0.0,7.0,0.0,surface,,1.0,,,,,False
u,m/s,U-component of wind,东西风,0.0,2.0,2.0,,,,,,,,False
u10m,m/s,V-component of wind at 10 meters height,10米东西风,0.0,2.0,2.0,heightAboveGround,10.0,103.0,10.0,,,instant,False
u10mmax#3,m/s,Maximum of u10m during 3 hours,3小时最大10米东西风,0.0,2.0,2.0,heightAboveGround,10.0,103.0,10.0,,,max,False
v,m/s,V-component of wind,南北风,0.0,2.0,3.0,,,,,,,,False
v10m,m/s,V-component of wind at 10 meters height,10米南北风,0.0,2.0,3.0,heightAboveGround,10.0,103.0,10.0,,,instant,False
v10mmax#3,m/s,Maximim of v10m during 3 hours,3小时最大10米南北风,0.0,2.0,3.0,heightAboveGround,10.0,103.0,10.0,,,max,False
w,m/s,Vertical velocity (geometric),等高面垂直速度,0.0,2.0,9.0,,,100.0,,,,,False
wmax#3,m/s,Maximum of w during 3 hours,3小时等高面最大垂直速度,0.0,2.0,9.0,,,100.0,,,,max,False
vor,s-1,Relative vorticity,涡度,0.0,2.0,12.0,,,,,,,,False
div,s-1,Relative divergence,散度,0.0,2.0,13.0,,,,,,,,False
q,kg/kg,Specific humidity,比湿,0.0,1.0,0.0,,,,,,,,False
q2m,kg/kg,Specific humidity at 2 meters height,2米比湿,0.0,1.0,0.0,heightAboveGround,2.0,103.0,2.0,,,,False
rh,%,Relative humidity,相对湿度,0.0,1.0,1.0,,,,,,,,False
rh2m,%,Relative humidity at 2 meters height,2米相对湿度,0.0,1.0,1.0,heightAboveGround,2.0,103.0,2.0,,,,False
pwat,kg/kg,Precipitable water,整层可降水量,0.0,1.0,3.0,atmosphere,,10.0,,,,,False
rain,mm,Total precipitation,总累计降水,0.0,1.0,8.0,surface,,1.0,,,,accum,False
rainnc,mm,Large scale precipitation,大尺度降水,0.0,1.0,9.0,surface,,1.0,,,,accum,False
rainc,mm,Convective precipitation,对流性降水,0.0,1.0,10.0,surface,,1.0,,,,accum,False
sf,m,Snow melt  ,降雪,0.0,1.0,29.0,surface,,1.0,,,,accum,False
snow,m,Snow melt  ,降雪,0.0,1.0,29.0,surface,,1.0,,,,accum,True
sd,m,Snow depth  ,雪深,0.0,1.0,11.0,surface,,1.0,,,,,False
tcc,%,Total cloud cover,总云量,0.0,6.0,1.0,entireAtmosphere,,,,,,,False
lcc,%,Low cloud cover,低云量,0.0,6.0,3.0,entireAtmosphere,,,,,,,False
mcc,%,Medium cloud cover,中云量,0.0,6.0,4.0,entireAtmosphere,,,,,,,False
hcc,%,High cloud cover,高云量,0.0,6.0,5.0,entireAtmosphere,,,,,,,False
bli,K,Best lifted index,最优抬升指数,0.0,7.0,1.0,surface,,,,,,,False
albedo,%,Albedo,反射率,0.0,19.0,1.0,surface,,1.0,,,,,False
st,K,Soil temperature,土壤温度,0.0,0.0,0.0,depthBelowLandLayer,,106.0,,,,,False
st(0-10),,,0-10cm土壤温度,0.0,0.0,0.0,depthBelowLandLayer,,106.0,0.0,106.0,0.1,,False
st(10-40),,,10-40cm土壤温度,0.0,0.0,0.0,depthBelowLandLayer,,106.0,0.1,106.0,0.4,,False
st(40-100),,,40-100cm土壤温度,0.0,0.0,0.0,depthBelowLandLayer,,106.0,0.4,106.0,1.0,,False
st(100-200),,,100-200cm土壤温度,0.0,0.0,0.0,depthBelowLandLayer,,106.0,1.0,106.0,2.0,,False
ro,m,Runoff,径流,0.0,1.0,227.0,,,,,,,,False
znt,m,roughness length,粗糙度,0.0,3.0,228.0,surface,,1.0,,,,,False
gsw,w/m2,Net short wave radiation flux,净短波辐射通量|净太阳辐射通量,0.0,4.0,9.0,,,,,,,,False
ssr,w/m2,Net short wave radiation flux,净短波辐射通量|净太阳辐射通量,0.0,4.0,9.0,,,,,,,,True
glw,w/m2,Net long wave radiation flux,净长波辐射通量|净热辐射通量,0.0,5.0,5.0,,,,,,,,False
str,w/m2,Net long wave radiation flux,净长波辐射通量|净热辐射通量,0.0,5.0,5.0,,,,,,,,True
gswc,w/m2,"Net short wave radiation flux, clear sky",晴空地面净短波辐射通量,0.0,4.0,11.0,surface,,1.0,,,,,False
tsw,w/m2,TOE net short wave radiation flux,大气顶净短波辐射通量,0.0,4.0,9.0,nominalTop,,8.0,,,,accum,False
tswc,w/m2,"TOE net short wave radiation flux, clear sky",晴空大气顶净短波辐射通量,0.0,4.0,11.0,nominalTop,,8.0,,,,,False
slhf,ws/m2,Latent heat net flux,潜热辐射,0.0,0.0,10.0,,,,,,,,False
sshf,ws/m2,Sensible heat net flux,感热辐射,2.0,0.0,24.0,,,,,,,,False
saunaidx,,Sauna index,桑拿指数,0.0,0.0,225.0,,,,,,,,False
kidx,C,K index,K指数,0.0,7.0,2.0,meanSea,,101.0,,,,,False
ki,C,K index,K指数,0.0,7.0,2.0,meanSea,,101.0,,,,,True
k,C,K index,K指数,0.0,7.0,2.0,meanSea,,101.0,,,,,True
qflx,1e-1.g/(cm.hpa.s),Moisture flux,水汽通量,0.0,1.0,224.0,,,,,,,,False
qdiv,1e-7.g/(cm2.hpa.s),Moisture flux divergence,水汽通量散度,0.0,1.0,225.0,,,,,,,,False
qfxsfc,kg/(m2.s),Surface moisture flux,地表水汽通量,0.0,1.0,229.0,surface,,1.0,,,,,False
qfx,kg/(m2.s),Surface moisture flux,地表水汽通量,0.0,1.0,229.0,surface,,1.0,,,,,True
sw,m3/m3,Soil moisture,土壤湿度,0.0,1.0,0.0,depthBelowLandLayer,,106.0,,,,,False
sw(0-10),,,0-10cm土壤湿度,0.0,1.0,0.0,depthBelowLandLayer,,106.0,0.0,106.0,0.1,,False
sw(10-40),,,10-40cm土壤湿度,0.0,1.0,0.0,depthBelowLandLayer,,106.0,0.1,106.0,0.3,,False
sw(40-100),,,40-100cm土壤湿度,0.0,1.0,0.0,depthBelowLandLayer,,106.0,0.3,106.0,0.6,,False
sw(100-200),,,100-200cm土壤湿度,0.0,1.0,0.0,depthBelowLandLayer,,106.0,0.6,106.0,1.0,,False
wess,N/m2,West-east surface wind stress,东西向风应力,0.0,2.0,227.0,surface,,1.0,,,,,False
nsss,N/m2,North-south surface wind stress,南北向风应力,0.0,2.0,228.0,surface,,1.0,,,,,False
sweatidx,,Sweat index,强天气胁迫指数,0.0,7.0,5.0,surface,,1.0,,,,,False
sweat,,Sweat index,强天气胁迫指数,0.0,7.0,5.0,surface,,1.0,,,,,True
qc,kg/kg,Cloud mixing ratio,云水混合比,0.0,1.0,22.0,,,,,,,,False
hfx,w/m2,Surface heat flux,地表热通量,0.0,0.0,24.0,surface,,1.0,,,,accum,False
cin,J/kg,Convective inhibition,对流抑制能量,0.0,7.0,7.0,surface,,1.0,,,,,False
cape,J/kg,Convective available potential energy,对流有效位能,0.0,7.0,6.0,surface,,1.0,,,,,False
pc,Pa,Condensation layer pressure,抬升凝结高度,0.0,1.0,228.0,surface,,1.0,,,,,False
rain5max,mm,5-minutes accumulated precipitation,5分钟最大降水,0.0,1.0,241.0,surface,,1.0,,,,max,False
qr,kg/kg,Rain mixing ratio,雨水混合比,0.0,1.0,24.0,,,,,,,,False
qs,kg/kg,Snow mixing ratio,雪水混合比,0.0,1.0,25.0,,,,,,,,False
qi,kg/kg,Ice mixing ratio,冰水混合比,0.0,1.0,23.0,,,,,,,,False
graupel,kg/kg,Graupel,霰,0.0,1.0,32.0,,,,,,,,False
precitype,,Precipitation type,降水相态,0.0,1.0,19.0,,,,,,,,False
mpv1,1e-6.k/(m2.s.kg),Moist potential vorticity(vertical part),湿位涡垂直分量,0.0,2.0,225.0,,,,,,,,False
mpv2,1e-6.k/(m2.s.kg),Moist potential vorticity(horizontal part),湿位涡水平分量,0.0,2.0,226.0,,,,,,,,False
shr,s-1,Vertical speed shear,垂直风切变,0.0,2.0,25.0,,,,,,,,False
shr(0-1000),,,0-1km垂直风切变,0.0,2.0,25.0,heightAboveGroundLayer,1000.0,103.0,1000.0,103.0,0.0,,False
shr(0-3000),,,0-3km垂直风切变,0.0,2.0,25.0,heightAboveGroundLayer,3000.0,103.0,3000.0,103.0,0.0,,False
shr(0-6000),,,0-6km垂直风切变,0.0,2.0,25.0,heightAboveGroundLayer,6000.0,103.0,6000.0,103.0,0.0,,False
shrmax,m/s,maximum vector of vertical wind shear,最大垂直风切变矢量,0.0,2.0,237.0,,,,,,,,False
shrmax(0-600),,,0-600m最大垂直风切变矢量,0.0,2.0,237.0,heightAboveGroundLayer,600.0,103.0,600.0,103.0,0.0,,False
shrmaxdir,degree,direction of maxium vertical wind shear,最大垂直风切变的方向,0.0,2.0,238.0,,,,,,,,False
shrmaxdir(0-600),,,0-600m最大垂直风切变的方向,0.0,2.0,238.0,heightAboveGroundLayer,600.0,103.0,600.0,103.0,0.0,,False
srh,J/kg,Storm relative helicity,风暴螺旋度,0.0,7.0,8.0,,,,,,,,False
srh(0-1000),,,0-1km垂直风暴螺旋度,0.0,7.0,8.0,heightAboveGroundLayer,1000.0,103.0,1000.0,103.0,0.0,,False
srh(0-3000),,,0-3km垂直风暴螺旋度,0.0,7.0,8.0,heightAboveGroundLayer,3000.0,103.0,1000.0,103.0,0.0,,False
uhmax,m2/s2,Updraught helicity,最大上升螺旋度,0.0,7.0,15.0,,,,,,,max,False
uhmax(2000-5000),,,2-5km最大上升螺旋度,0.0,7.0,15.0,heightAboveGroundLayer,5000.0,103.0,5000.0,103.0,2000.0,max,False
src,,Skin reservoir content  ,表面蓄水池含量,0.0,1.0,226.0,,,,,,,,False
pblh,m,Planet boundary layer height,边界层高度,0.0,3.0,18.0,surface,,1.0,,,,,False
gusw,w/m2,Surface upward short-wave radiation flux(surface),地面向上短波辐射通量,0.0,4.0,8.0,surface,,1.0,,,,,False
gswu,w/m2,Surface upward short-wave radiation flux(surface),地面向上短波辐射通量,0.0,4.0,8.0,surface,,1.0,,,,,True
gulw,w/m2,Surface upward long-wave radiation flux(surface),地面向上长波辐射通量,0.0,5.0,4.0,surface,,1.0,,,,,False
gslu,w/m2,Surface upward long-wave radiation flux(surface),地面向上长波辐射通量,0.0,5.0,4.0,surface,,1.0,,,,,True
tusw,w/m2,Upward short-wave radiation flux(top of atmoshpere),大气顶向上（外）短波辐射通量,0.0,4.0,8.0,nominalTop,,8.0,,,,,False
tulw,w/m2,Upward long-wave radiation flux(top of atmoshpere),大气顶向上（外）长波辐射通量,0.0,5.0,4.0,nominalTop,,8.0,,,,,False
gdlwc,w/m2,"Surface downward long-wave radiation flux, clear sky",地面晴空向下长波辐射通量,0.0,5.0,8.0,surface,,1.0,,,,,False
guswc,w/m2 S,"urface upward short-wave radiation flux, clear sky",地面晴空向上短波辐射通量,0.0,4.0,53.0,surface,,1.0,,,,,False
gulwc,w/m2,"Surface upward long-wave radiation flux, clear sky",地面晴空向上长波辐射通量,0.0,5.0,224.0,surface,,1.0,,,,,False
tuswc,w/m2,"TOE upward short-wave radiation flux, clear sky",大气顶晴空向上短波辐射通量,0.0,4.0,53.0,nominalTop,,8.0,,,,,False
tulwc,w/m2,"TOE upward long-wave radiation flux, clear sky",大气顶晴空向上长波辐射通量,0.0,5.0,224.0,nominalTop,,8.0,,,,,False
btv,K,Radiation bright temperature(vapor channel),辐射亮温（水汽通道）,0.0,4.0,227.0,surface,,,,,,,False
bti,K,Radiation bright temperature(infrared channel),辐射亮温（红外通道）,0.0,4.0,228.0,surface,,,,,,,False
tadv,1.e-6 k/s,Temperature advection,温度平流,0.0,0.0,224.0,,,,,,,,False
t_adv,1.e-6 k/s,Temperature advection,温度平流,0.0,0.0,224.0,,,,,,,,True
tvw,kg/m2,Total column integrated water vapor,总柱状比湿,0.0,1.0,64.0,atmosphere,,10.0,,,,,False
tcw,kg/m2,Total column integrated cloud water,总柱状云水,0.0,1.0,69.0,atmosphere,,10.0,,,,,False
tiw,kg/m2,Total column integrated cloud ice,总柱状云冰,0.0,1.0,70.0,atmosphere,,10.0,,,,,False
voradv,1.e-11/(s2),Relative vorticity advection,涡度平流,0.0,2.0,224.0,,,,,,,,False
vor_adv,1.e-11/(s2),Relative vorticity advection,涡度平流,0.0,2.0,224.0,,,,,,,,True
cc,%,Cloud cover,云量,0.0,6.0,22.0,,,,,,,,False
pip,,Perturbed Exner pressure,Exner气压扰动,0.0,3.0,224.0,,,,,,,,False
thp,,Perturbed potential temperature,位温扰动,0.0,0.0,229.0,,,,,,,,False
ri,,Richardson number,理查森数,0.0,7.0,12.0,,,,,,,,False
risfc,,Richardson number of surface layer,近地面总体理查森数,0.0,7.0,12.0,surface,,1.0,,,,,False
ripbl,,Richardson number of PBL,边界层总体理查森数,0.0,7.0,12.0,,,166.0,,,,,False
qpe,mm,Quantitive precipitation estimate,定量降水估计,0.0,1.0,230.0,,,,,,,,False
rhmax,%,Maximum of rh ,最大相对湿度,0.0,1.0,231.0,,,,,,,,False
rhmax2m,%,Maximum of rh at 2 meters height,2米最大相对湿度,0.0,1.0,231.0,heightAboveGround,2.0,103.0,2.0,,,,False
rhmin,%,Minimum of rh,最小相对湿度,0.0,1.0,232.0,,,,,,,,False
rhmin2m,%,Minimum of rh at 2 meters height,2米最小相对湿度,0.0,1.0,232.0,heightAboveGround,2.0,103.0,2.0,,,,False
rainp,%,Probability of precipitation,降水概率,0.0,1.0,233.0,,,,,,,,False
rainpr,mm,Process precipitation,过程降水量,0.0,1.0,234.0,,,,,,,,False
rainh,mm,Short time heavy precipitation,短时强降水,0.0,1.0,235.0,,,,,,,,False
hail,mm,Hail,冰雹,0.0,1.0,236.0,,,,,,,,False
raint,mm,Precipitation of typhoon,台风降水,0.0,1.0,237.0,,,,,,,,False
light,,Lightning,闪电,0.0,1.0,238.0,,,,,,,,False
htdn,,Number of high temperature days ,高温日数,0.0,0.0,226.0,,,,,,,,False
ata,K,Average temperature anomaly,平均气温距平,0.0,0.0,227.0,,,,,,,,False
tn,K,Minimum of air temperature for NMC medium-term prediction only,最低气温(仅限NMC使用),0.0,0.0,228.0,,,,,,,,False
thunwind,,Thunderstorm wind,雷暴大风,0.0,2.0,229.0,,,,,,,,False
seawave,m,Sea wave,海浪,0.0,2.0,230.0,,,,,,,,False
weph,,Weather phenomenon,天气现象,0.0,2.0,231.0,,,,,,,,False
haze,,Haze,霾,0.0,2.0,232.0,,,,,,,,False
apd,,Condition of air pollution,空气污染条件,0.0,2.0,233.0,,,,,,,,False
fog,,Fog,雾,0.0,2.0,234.0,,,,,,,,False
cdbz,dbz,Composite radar reflectivity,雷达组合反射率,0.0,16.0,224.0,surface,,,,,,,False
cr,dbz,Composite radar reflectivity,雷达组合反射率,0.0,16.0,224.0,surface,,,,,,,True
c_dbz,dbz,Composite radar reflectivity,雷达组合反射率,0.0,16.0,224.0,surface,,,,,,,True
cdbzmax#3,dbz,Maximum of cdbz during 3 hours,3小时最大雷达组合反射率,0.0,16.0,224.0,surface,,,,,,max,False
cr#3,dbz,Maximum of cdbz during 3 hours,3小时最大雷达组合反射率,0.0,16.0,224.0,surface,,,,,,max,True
dbz,dbz,Radar reflectivity,雷达反射率,0.0,16.0,225.0,,,,,,,,False
dbzmax#3,dbz,Maximum of dbz during 3 hours,3小时最大雷达反射率,0.0,16.0,225.0,,,,,,,max,False
dbzmaxh,m,height of maximum of radar reflectivity,最大雷达回波的高度,0.0,16.0,226.0,,,,,,,,False
vi,m2/s,Ventilation Index,通风系数,0.0,2.0,235.0,,,,,,,,False
ht0,m,Height of 0 degree isothermal level,0度层高度,0.0,3.0,225.0,surface,,1.0,,,,,False
ht2,m,Height of -20 degree isothermal level,-20度层高度,0.0,3.0,226.0,surface,,1.0,,,,,False
tti,K,Total totals index,总指数,0.0,7.0,4.0,surface,,1.0,,,,,False
wi,,Wind index,大风指数,0.0,2.0,236.0,surface,,1.0,,,,,False
hi,,Hail index ,冰雹指数,0.0,1.0,239.0,surface,,1.0,,,,,False
hib,m,Height of inversion base ,逆温层底高度,0.0,3.0,227.0,,,,,,,,False
II,oC/hm,Inversion intensity,逆温强度,0.0,1.0,240.0,,,,,,,,False
dcape,J/kg,Down convective available potential energy,下沉有效位能,0.0,7.0,224.0,,,,,,,,False
ssi,,Storm strength index,风暴强度指数,0.0,7.0,225.0,,,,,,,,False
si,K,Showalter index,沙氏指数,0.0,7.0,13.0,surface,,1.0,,,,,False
asi,,Atmospheric stability index,大气稳定度指数,0.0,7.0,226.0,,,,,,,,False
gust,m/s,Wind speed (gust),,0.0,2.0,22.0,,,,,,,,False
cldt,m,cloud top,,0.0,6.0,12.0,,,,,,,,False
cldb,m,cloud base,,0.0,6.0,11.0,,,,,,,,False
ceiling,m,ceiling,,0.0,6.0,13.0,,,,,,,,False
"""


def _get_cemc_param_table() -> pd.DataFrame:
    f = io.StringIO(_CEMC_PARAM_TABLE_CONTENT)
    df = pd.read_table(
        f,
        header=0,
        sep=",",
    )
    return df


CEMC_PARAM_TABLE = _get_cemc_param_table()


import argparse
from typing import Union

import click
import pandas as pd

from ._config import (
    get_default_local_config_path,
    find_config,
    load_config,
)
from ._util import find_file


def main():
    cli()


def print_local_help(ctx, param, value):
    if value is False:
        return
    click.echo(ctx.get_help())

    click.echo("\nDifferent steams use different additional options.\n")

    oper_parser = create_oper_option_parser()
    click.echo(oper_parser.format_help())

    eps_parser = create_eps_option_parser()
    click.echo(eps_parser.format_help())

    ctx.exit()


@click.group()
def cli():
    pass


@cli.command("local", context_settings=dict(
    ignore_unknown_options=True,
))
@click.option("--data-type", required=True, help="data type, such as cma_gfs_gmf/grib2/orig")
@click.option("--data-level", default="archive,storage", help="data level, split by comma, such as archive,storage")
@click.option("--data-class", default="od", help="data class, such as od for operation data, or cm for common data.")
@click.option("--config-dir", default=None, help="config directory")
@click.option(
    "--help", "-h",
    is_flag=True,
    expose_value=False,
    is_eager=True,
    callback=print_local_help,
    help="Show this message and exit.")
@click.argument('query_args', nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def find_local(ctx, data_type, data_level, data_class, config_dir, query_args):
    if config_dir is None:
        config_dir = get_default_local_config_path()

    config_file_path = find_config(config_dir, data_type, data_class)
    if config_file_path is None:
        raise ValueError(f"data type is not found: {data_type}")

    if data_level == "":
        data_level = None
    else:
        data_level = data_level.split(",")

    config = load_config(config_file_path)

    stream = config["query"]["stream"]

    if stream in ("oper", "cm"):
        find_oper_file(config, data_level, query_args)
    elif stream == "eps":
        find_eps_file(config, data_level, query_args)
    else:
        raise ValueError(f"stream type is not supported: {stream}")


def find_oper_file(config: dict, data_level: Union[str, list[str]], query_args: tuple):
    parser = create_oper_option_parser()
    args = parser.parse_args(query_args)

    forecast_time = pd.to_timedelta(args.forecast_time)
    start_time = pd.to_datetime(args.start_time, format="%Y%m%d%H")

    file_path = find_file(config, data_level, start_time, forecast_time)
    if file_path is None:
        print("None")
    else:
        print(file_path)


def find_eps_file(config: dict, data_level: Union[str, list[str]], query_args: tuple):
    parser = create_eps_option_parser()
    args = parser.parse_args(query_args)

    forecast_time = pd.to_timedelta(args.forecast_time)
    start_time = pd.to_datetime(args.start_time, format="%Y%m%d%H")

    file_path = find_file(config, data_level, start_time, forecast_time, number=args.number)
    if file_path is None:
        print("None")
    else:
        print(file_path)


def create_oper_option_parser():
    parser = argparse.ArgumentParser(
        description='Additional options for stream oper.',
        usage=None,
        add_help=False
    )
    parser.add_argument(
        '--start-time',
        dest="start_time",
        help='start time, such as YYYMMDDHH'
    )
    parser.add_argument(
        '--forecast-time',
        dest='forecast_time',
        default="0h",
        help='forecast time, such as 3h'
    )
    return parser


def create_eps_option_parser():
    parser = argparse.ArgumentParser(
        description='Additional options for stream eps.',
        usage=None,
        add_help=False
    )
    parser.add_argument(
        '--start-time',
        dest="start_time",
        help='start time, such as YYYMMDDHH'
    )
    parser.add_argument(
        '--forecast-time',
        dest='forecast_time',
        default="0h",
        help='forecast time, such as 3h'
    )
    parser.add_argument(
        '--number',
        dest='number',
        type=int,
        help='member number')
    return parser


# copy from argparse module
def _format_help(parser: argparse.ArgumentParser):
    """
    Notes
    -----
    This is an experimental function. Do not use it.
    """
    formatter = parser._get_formatter()

    # description
    formatter.add_text(parser.description)

    # positionals, optionals and user-defined groups
    for action_group in parser._action_groups:
        formatter.start_section(action_group.title)
        formatter.add_text(action_group.description)
        formatter.add_arguments(action_group._group_actions)
        formatter.end_section()

    # epilog
    # formatter.add_text(parser.epilog)

    # determine help from format above
    return formatter.format_help()


if __name__ == "__main__":
    cli()



from typing import Dict, Optional, Union, Tuple
import math

import numpy as np
import xarray as xr
import pandas as pd

import eccodes

from reki.format.grib.common import MISSING_VALUE

# from loguru import logger


def create_data_array_from_message(
        message,
        level_dim_name: Optional[str] = None,
        field_name: Optional[str] = None,
        missing_value: Optional[float] = None,
        fill_missing_value: Optional = np.nan,
        values: Optional[np.ndarray] = None,
) -> xr.DataArray:
    """
    Create ``xarray.DataArray`` from one GRIB2 message.

    Parameters
    ----------
    message
        grib message id loaded by ecCodes python API.
    level_dim_name
    field_name
    missing_value
        set missingValue key in GRIB message before get array.
        If set None, use MISSING_VALUE set in common module.
        NOTE: ecCodes use 9999 as default missing value.
    fill_missing_value
        filled value to replace missing value point in array.
        default is np.nan.
        If set None, missing value will not be changed.
    values
        message values. if None, function will decode values from message.
        if set, function will use values instead of decode message.
    """
    if missing_value is None:
        missing_value = MISSING_VALUE
    eccodes.codes_set(message, "missingValue", missing_value)

    if values is None:
        # logger.info("decoding...")
        values = eccodes.codes_get_double_array(message, "values")
        # logger.info("decoding...done")

    if fill_missing_value is not None:
        np.place(values, values == missing_value, fill_missing_value)

    attr_keys = [
        'edition',
        'centre',
        'subCentre',
        'tablesVersion',
        "localTablesVersion",
        'dataType',
        'dataDate',
        'dataTime',
        'validityDate',
        'validityTime',
        'step',
        'stepType',
        'stepUnits',
        'stepRange',
        'endStep:int',
        'count'
    ]

    parameter_keys = [
        "name",
        "shortName",
        'cfName',
        'discipline',
        'parameterCategory',
        'parameterNumber',
        'units',
    ]

    grid_keys = [
        'gridType',
        'gridDefinitionDescription',
        'numberOfPoints',
        "missingValue",
        'latitudeOfFirstGridPointInDegrees',
        'longitudeOfFirstGridPointInDegrees',
        'latitudeOfLastGridPointInDegrees',
        'longitudeOfLastGridPointInDegrees',
        'iDirectionIncrementInDegrees',
        'jDirectionIncrementInDegrees',
        'Ni',
        'Nj',
    ]

    level_keys = [
        'typeOfLevel',
        'level',
        "typeOfFirstFixedSurface",
        "typeOfSecondFixedSurface",
        "scaleFactorOfFirstFixedSurface",
        "scaledValueOfFirstFixedSurface",
        "scaleFactorOfSecondFixedSurface",
        "scaledValueOfSecondFixedSurface",
    ]

    all_keys = attr_keys + parameter_keys + grid_keys + level_keys

    all_attrs = {}
    key_type_mapper = {
        "int": int,
        "float": float,
        "str": str,
    }
    for key in all_keys:
        tokens = key.split(":")
        if len(tokens) == 1:
            try:
                value = eccodes.codes_get(message, key)
            except:
                value = "undef"
        elif len(tokens) == 2:
            key_name = tokens[0]
            key_type = key_type_mapper[tokens[1]]
            try:
                value = eccodes.codes_get(message, key_name, key_type)
            except:
                value = "undef"
        else:
            value = "undef"
        all_attrs[key] = value

    latitude_of_first_grid_point_in_degrees = all_attrs["latitudeOfFirstGridPointInDegrees"]
    longitude_of_first_grid_point_in_degrees = all_attrs["longitudeOfFirstGridPointInDegrees"]
    latitude_of_last_grid_point_in_degrees = all_attrs["latitudeOfLastGridPointInDegrees"]
    longitude_of_last_grid_point_in_degrees = all_attrs["longitudeOfLastGridPointInDegrees"]
    ni = all_attrs["Ni"]
    nj = all_attrs["Nj"]

    values = values.reshape(nj, ni)
    lons = np.linspace(
        longitude_of_first_grid_point_in_degrees, longitude_of_last_grid_point_in_degrees, ni,
        endpoint=True
    )
    lats = np.linspace(
        latitude_of_first_grid_point_in_degrees, latitude_of_last_grid_point_in_degrees, nj,
        endpoint=True
    )

    # coords
    coords = {}

    # add time and step coordinate
    time_name, value = get_time_from_attrs(all_attrs)
    coords[time_name] = value

    step_name, value = get_step_from_attrs(all_attrs)
    coords[step_name] = value

    # add valid time coordinate
    valid_time_name, value = get_valid_time_from_attrs(all_attrs)
    if valid_time_name is not None:
        coords[valid_time_name] = value

    # add level coordinate
    level_name, value = get_level_from_attrs(all_attrs, level_dim_name)
    coords[level_name] = value

    coords["latitude"] = xr.Variable(
        "latitude",
        lats,
        attrs={
            "units": "degrees_north",
            "standard_name": "latitude",
            "long_name": "latitude"
        },
    )
    coords["longitude"] = xr.Variable(
        "longitude",
        lons,
        attrs={
            "units": "degrees_east",
            "standard_name": "longitude",
            "long_name": "longitude"
        }
    )

    #   check ENS
    key_name = "perturbationNumber"
    try:
        value = eccodes.codes_get(message, key_name)
    except:
        value = None
    if value is not None:
        coords["number"] = value

    dims = ("latitude", "longitude")

    data_attrs = {f"GRIB_{key}": all_attrs[key] for key in attr_keys if all_attrs[key] not in ("undef", "unknown")}

    # set long_name
    if "GRIB_name" in data_attrs:
        data_attrs["long_name"] = data_attrs["GRIB_name"]
    else:
        name = (f"discipline={all_attrs['discipline']} "
                f"parmcat={all_attrs['parameterCategory']} "
                f"parm={all_attrs['parameterNumber']}")
        data_attrs["long_name"] = name

    # set name
    if field_name is not None:
        var_name = field_name
    elif "shortName" in all_attrs and all_attrs["shortName"] != "unknown":
        var_name = all_attrs["shortName"]
    else:
        var_name = f"{all_attrs['discipline']}_{all_attrs['parameterCategory']}_{all_attrs['parameterNumber']}"

    # set units
    if "GRIB_units" in data_attrs:
        data_attrs["units"] = data_attrs["GRIB_units"]

    data = xr.DataArray(
        values,
        dims=dims,
        coords=coords,
        attrs=data_attrs,
        name=var_name,
    )

    return data


def get_level_coordinate_name(data: xr.DataArray) -> Optional[str]:
    """
    Get coordinate name from ``xarray.DataArray`` object.

    NOTE: please use typeOfLevel if available and don't use this function.

    Parameters
    ----------
    data: xr.DataArray

    Returns
    -------
    str or None
    """
    coords = data.coords
    for coord in coords:
        if coord.startswith("level_"):
            return coord
    return None


def get_time_from_attrs(all_attrs):
    start_time = pd.to_datetime(
        f"{all_attrs['dataDate']}{all_attrs['dataTime']:04}",
        format="%Y%m%d%H%M"
    )
    return "time", start_time


def get_step_from_attrs(all_attrs):
    if all_attrs["stepUnits"] == 1:
        forecast_hour = pd.Timedelta(hours=all_attrs["endStep:int"])
    elif all_attrs["stepUnits"] == 0:
        forecast_hour = pd.Timedelta(minutes=all_attrs["endStep:int"])
    elif all_attrs["stepUnits"] == 2:
        forecast_hour = pd.Timedelta(days=all_attrs["endStep:int"])
    else:
        raise ValueError(f"stepUnits is not supported: {all_attrs['stepUnits']}")
    return "step", forecast_hour


def get_valid_time_from_attrs(all_attrs):
    if all_attrs["validityDate"] in ("undef", "unknown"):
        return None, None
    if all_attrs["validityTime"] in ("undef", "unknown"):
        return None, None
    valid_time = pd.to_datetime(f"{all_attrs['validityDate']}{all_attrs['validityTime']:04}")
    return "valid_time", valid_time


def get_level_from_attrs(
        all_attrs: Dict,
        level_dim_name: Optional[str] = None,
) -> Tuple[str, Union[float, int]]:
    """
    Get level coordinate name and value.

    If message has typeOfLevel, use typeOfLevel as coordinate name,
    else use "level_{typeOfFirstFixedSurface}",
    or "level_{typeOfFirstFixedSurface}_{typeOfSecondFixedSurface}" if typeOfSecondFixedSurface is not 255.

    Parameters
    ----------
    all_attrs
    level_dim_name

    Returns
    -------
    typing.Tuple[str, float or int]

    """
    if level_dim_name == "isobaricInPa":
        value = math.pow(10, all_attrs["scaleFactorOfFirstFixedSurface"]) * all_attrs["scaledValueOfFirstFixedSurface"]
        return level_dim_name, value
    elif level_dim_name in ("isobaricInhPa", "pl"):
        value = math.pow(10, all_attrs["scaleFactorOfFirstFixedSurface"]) * all_attrs["scaledValueOfFirstFixedSurface"]
        return level_dim_name, value / 100.0
    elif isinstance(level_dim_name, str):
        # TODO: add check for level_type="pl"
        return level_dim_name, all_attrs["level"]
    elif level_dim_name is None:
        if all_attrs["typeOfLevel"] not in ("undef", "unknown"):
            return all_attrs["typeOfLevel"], all_attrs["level"]
        else:
            level_name = f"level_{all_attrs['typeOfFirstFixedSurface']}"
            if all_attrs['typeOfSecondFixedSurface'] != 255:
                level_name += f"{all_attrs['typeOfSecondFixedSurface']}"
            return level_name, all_attrs["level"]
    else:
        raise TypeError(f"level_dim_name is not supported: {level_dim_name}")



import typing
from typing import Union, List, Dict, Optional
from pathlib import Path

import eccodes
import xarray as xr
from tqdm import tqdm

from ._level import _fix_level
from ._util import _check_message
from ._xarray import create_data_array_from_message, get_level_coordinate_name
from reki._util import _load_first_variable
from reki.format.grib.common._parameter import convert_parameter


def load_field_from_file(
        file_path: Union[str, Path],
        parameter: Union[str, Dict] = None,
        level_type: Union[str, Dict] = None,
        level: Union[int, float, List, Dict, str] = None,
        level_dim: Optional[str] = None,
        field_name: Optional[str] = None,
        show_progress: bool = False,
        **kwargs
) -> Optional[xr.DataArray]:
    """
    Load **one** field from local GRIB2 file using eccodes-python.
    Or load multi levels into one field.

    Parameters
    ----------
    file_path : str or Path
    parameter : str or typing.Dict
        parameter name.

        Use GRIB key `shortName` or a dict of filter conditions such as:

        .. code-block::

            {
                "discipline": 0,
                "parameterCategory": 2,
                "parameterNumber": 225,
            }

    level_type : str or typing.Dict or None
        level type.

        - Use "pl", "ml" or "sfc". They will be converted into dict.
        - Use GRIB key `typeOfLevel`, such as
            - "isobaricInhPa"
            - "isobaricInPa"
            - "surface"
            - "heightAboveGround"
            - "nominalTop"
            - "atmosphere"
            - "meanSea"
            - "depthBelowLandLayer"
            - ...
          See https://apps.ecmwf.int/codes/grib/format/edition-independent/3/ for more values.
        - If `typeOfLevel` is not available, use dict to specify filter conditions.
          For example, to get one filed from GRAPES GFS modelvar GRIB2 file, use:

          .. code-block::

            {
                "typeOfFirstFixedSurface": 131
            }

    level : int or float or typing.List or typing.Dict or None
        level value(s).

        - If use a scalar, level will be a non-dimension coordinate.
        - If your want to extract multi levels, use a list and level will be a dimension (level, lat, lon).
        - If use a dict, message will be filtered by dict keys. Support custom calculate keys:
            - ``first_level``
            - ``second_level``
        - If use `"all"`, all levels of level_type will be packed in the result field.
        - If use `None`, only the first field will be returned.

    level_dim : str or None
        name of level dimension.
        If none, function will generate a name for level dim.
        If `level_type="pl"`, some values can be used:

            - `None` or `pl` or `isobaricInhPa`: level_dim is a float number with unit hPa.
            - `isobaricInPa`: level_dim is a float number with unit Pa.

    field_name :
        name of field.

        * If ``None``, generate field name automatically.
            * use ``parameter`` if it's a string
            * use GRIB Key shortName if it's not `unknown`
            * use GRIB Keys to combine a name: `discipline_parameterCategory_parameterNumber`
        * If string, set the field's name with it.

    show_progress : bool
        show progress bar.

    Returns
    -------
    DataArray or None:
        DataArray if found, or None if not.

    Examples
    --------
    Load 850hPa temperature field from a GRIB2 file generated by GRAPES GFS.

    .. code-block:: pycon

        >>> load_field_from_file(
        ...     file_path="/sstorage1/COMMONDATA/OPER/NWPC/GRAPES_GFS_GMF/Prod-grib/2020031721/ORIG/gmf.gra.2020031800105.grb2",
        ...     parameter="t",
        ...     level_type="isobaricInhPa",
        ...     level=850,
        ... )
        <xarray.DataArray 't' (latitude: 720, longitude: 1440)>
        array([[249.19234375, 249.16234375, 249.16234375, ..., 249.15234375,
                249.19234375, 249.14234375],
               [249.45234375, 249.45234375, 249.42234375, ..., 249.45234375,
                249.44234375, 249.44234375],
               [249.69234375, 249.68234375, 249.68234375, ..., 249.70234375,
                249.67234375, 249.68234375],
               ...,
               [235.33234375, 235.45234375, 235.62234375, ..., 235.47234375,
                235.63234375, 235.48234375],
               [235.78234375, 235.91234375, 235.64234375, ..., 235.80234375,
                235.72234375, 235.82234375],
               [235.66234375, 235.86234375, 235.82234375, ..., 235.85234375,
                235.68234375, 235.70234375]])
        Coordinates:
            time           datetime64[ns] 2020-03-18
            step           timedelta64[ns] 4 days 09:00:00
            valid_time     datetime64[ns] 2020-03-22T09:00:00
            isobaricInhPa  int64 850
          * latitude       (latitude) float64 89.88 89.62 89.38 ... -89.38 -89.62 -89.88
          * longitude      (longitude) float64 0.0 0.25 0.5 0.75 ... 359.2 359.5 359.8
        Attributes:
            GRIB_edition:                    2
            GRIB_centre:                     babj
            GRIB_subCentre:                  0
            GRIB_tablesVersion:              4
            GRIB_localTablesVersion:         1
            GRIB_dataType:                   fc
            GRIB_dataDate:                   20200318
            GRIB_dataTime:                   0
            GRIB_validityDate:               20200322
            GRIB_validityTime:               900
            GRIB_step:                       105
            GRIB_stepType:                   instant
            GRIB_stepUnits:                  1
            GRIB_stepRange:                  105
            GRIB_endStep:                    105
            GRIB_name:                       Temperature
            GRIB_shortName:                  t
            GRIB_cfName:                     air_temperature
            GRIB_discipline:                 0
            GRIB_parameterCategory:          0
            GRIB_parameterNumber:            0
            GRIB_gridType:                   regular_ll
            GRIB_gridDefinitionDescription:  Latitude/longitude
            GRIB_typeOfFirstFixedSurface:    pl
            GRIB_typeOfLevel:                isobaricInhPa
            GRIB_level:                      850
            GRIB_numberOfPoints:             1036800
            GRIB_missingValue:               9999
            GRIB_units:                      K
            long_name:                       Temperature
            units:                           K

    """
    messages = []

    fixed_level_type, fixed_level_dim = _fix_level(level_type, level_dim)

    if field_name is None and isinstance(parameter, str):
        field_name = parameter

    parameter = convert_parameter(parameter)

    if show_progress:
        with open(file_path, "rb") as f:
            total_count = eccodes.codes_count_in_file(f)

    with open(file_path, "rb") as f:
        if show_progress:
            pbar = tqdm(
                total=total_count,
                desc="Filtering",
            )
        while True:
            message_id = eccodes.codes_grib_new_from_file(f)
            if message_id is None:
                break
            if show_progress:
                pbar.update(1)
            if not _check_message(message_id, parameter, fixed_level_type, level, **kwargs):
                eccodes.codes_release(message_id)
                continue
            messages.append(message_id)
            if isinstance(level, typing.List) or level == "all":
                continue
            else:
                break
        if show_progress:
            pbar.close()

    if len(messages) == 0:
        return None

    if len(messages) == 1:
        message_id = messages[0]
        data = create_data_array_from_message(
            message_id,
            level_dim_name=fixed_level_dim,
            field_name=field_name,
        )
        eccodes.codes_release(message_id)
        return data

    if len(messages) > 1:
        if show_progress:
            pbar = tqdm(
                total=len(messages),
                desc="Decoding",
            )

        def creat_array(message):
            array = create_data_array_from_message(
                message,
                level_dim_name=fixed_level_dim,
                field_name=field_name
            )
            if show_progress:
                pbar.update(1)
            return array

        xarray_messages = [creat_array(message) for message in messages]
        for m in messages:
            eccodes.codes_release(m)
        if show_progress:
            pbar.close()

        if level_dim is None:
            if isinstance(level_type, str):
                level_dim_name = level_type
            elif isinstance(level_type, typing.Dict):
                level_dim_name = get_level_coordinate_name(xarray_messages[0])
            else:
                raise ValueError(f"level_type is not supported: {level_type}")
        elif isinstance(level_dim, str):
            level_dim_name = level_dim
        else:
            raise ValueError(f"level_type is not supported: {level_type}")

        if show_progress:
            print("Packing...")

        data = xr.concat(xarray_messages, level_dim_name)
        return data

    return None


def load_field_from_files(
        file_list: List,
        parameter: Union[str, Dict],
        level_type: Union[str, Dict],
        level: Optional[Union[int, float, List, Dict]],
        level_dim: Optional[str] = None,
        show_progress: bool = False,
        **kwargs
) -> Optional[xr.DataArray]:
    """
    Load one field from multiply files.

    Parameters
    ----------
    file_list: typing.List
        file list.
    parameter: str or typing.Dict
        see ``load_field_from_file``
    level_type: str or typing.Dict
        see ``load_field_from_file``
    level: int or float or typing.List or None
        see ``load_field_from_file``
    level_dim: str or None
        level dimension name.
    show_progress: bool
        see ``load_field_from_file``

    Returns
    -------
    xr.DataArray or None:
        xr.DataArray if found, or None if not.

    """
    field_list = []
    for file_path in file_list:
        print(file_path)
        field = load_field_from_file(
            file_path,
            parameter=parameter,
            level_type=level_type,
            level=level,
            level_dim=level_dim,
            show_progress=show_progress,
            **kwargs
        )
        field_list.append(field)

    data_set = xr.combine_by_coords(
        [f.expand_dims(["time", "step"]).to_dataset() for f in field_list]
    )
    data = _load_first_variable(data_set)
    data = data.transpose("time", "step", ...)
    return data
